/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls/cpp/search.hpp
 *
 * Purpose:     recls C++ mapping - Search class.
 *
 * Created:     18th August 2003
 * Updated:     27th December 2005
 *
 * Home:        http://recls.org/
 *
 * Copyright (c) 2003-2005, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef RECLS_INCL_RECLS_CPP_HPP_SEARCH
#define RECLS_INCL_RECLS_CPP_HPP_SEARCH

/* File version */
#ifndef RECLS_DOCUMENTATION_SKIP_SECTION
# define RECLS_VER_RECLS_CPP_HPP_SEARCH_MAJOR       3
# define RECLS_VER_RECLS_CPP_HPP_SEARCH_MINOR       0
# define RECLS_VER_RECLS_CPP_HPP_SEARCH_REVISION    2
# define RECLS_VER_RECLS_CPP_HPP_SEARCH_EDIT        10
#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */

/** \file recls/cpp/search.hpp Definition of the \link recls::cpp::Search Search\endlink class, part of 
 * the \ref group_recls_cpp "recls C++ mapping"
 */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#include <recls/cpp/recls.hpp>      // reclspp root header
#include <recls/cpp/fileentry.hpp>  // reclspp::FileEntry
#include <recls/assert.h>           // RECLS_ASSERT
#include <utility>                  // pair

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{

namespace cpp
{
#endif /* !RECLS_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

/// This class provides functionality for executing a file search
///
/// \ingroup group_recls_cpp
class Search
{
/// \name Types
public:
/// @{
    typedef Search      class_type;
/// @}

/// \name Construction
/// @{
protected:
    typedef std::pair<hrecls_t, recls_rc_t> ctor_args_type;

    /// Creates a search for the items identified by \c pattern in the directory \c rootDir, according to the given \c flags.
    ///
    /// \param args The arguments pair, containing the recls search handle and the error code associated with that handle
    Search(ctor_args_type args);
public:
    /// Closes the search and releases resources
    ///
    /// \note This is made pure-virtual just to make the class abstract. It imposes a 
    /// tedious constraint on the inherited classes, but the alternative is risk this
    /// class being instantiated (if someone used the evil access #define trick).
    virtual ~Search() = 0;
/// @}

/// \name Operations
/// @{
public:
    /// Advances the search to the next position
    recls_rc_t      GetNext();
/// @}

/// \name Attributes
/// @{
public:
    /// Returns non-zero if there is more data available
    recls_bool_t    HasMoreElements() const;
    /// Returns the current entry
    ///
    /// \note The behaviour is undefined when the \c HasMoreElements() returns zero
    FileEntry       GetCurrentEntry() const;

    /// Returns the last error status code for this search object
    recls_rc_t      GetLastError() const;

    /// Returns the number of outstanding file-entry blocks associated with this search object
    recls_uint32_t  GetNumOutstandingDetails() const;
/// @}

// Implementation
protected:
    // This method is needed because there is no way to call a function in the 
    // MIL of a derived class and pass more than one argument to a base ctor
    recls_rc_t          &GetLastError_();

    static FileEntry    make_entry_(recls_info_t info);

// Members
private:
    hrecls_t /* const */    m_hSrch;
    mutable recls_rc_t      m_lastError;

// Not to be implemented
private:
    Search(class_type const &);
    class_type &operator =(class_type const &);
};

/* /////////////////////////////////////////////////////////////////////////////
 * Implementation
 */

inline /* static */ FileEntry   Search::make_entry_(recls_info_t info)
{
    return FileEntry(info);
}


inline Search::Search(Search::ctor_args_type args)
    : m_hSrch(args.first)
    , m_lastError(args.second)
{
    RECLS_ASSERT(NULL != m_hSrch || RECLS_FAILED(m_lastError));
    RECLS_ASSERT(NULL == m_hSrch || RECLS_SUCCEEDED(m_lastError));
}

inline Search::~Search()
{
    if(NULL != m_hSrch)
    {
        Recls_SearchClose(m_hSrch);
    }
}

inline recls_rc_t Search::GetNext()
{
    RECLS_ASSERT(NULL != m_hSrch);

    recls_rc_t  rc  =   Recls_GetNext(m_hSrch);

    if(RECLS_FAILED(rc))
    {
        Recls_SearchClose(m_hSrch);
        m_hSrch = NULL;
    }

    m_lastError = rc;

    return rc;
}

inline recls_bool_t Search::HasMoreElements() const
{
    return RECLS_SUCCEEDED(m_lastError);
}

inline FileEntry Search::GetCurrentEntry() const
{
    recls_info_t    info;
    recls_rc_t      rc  =   Recls_GetDetails(m_hSrch, &info);

    return RECLS_SUCCEEDED(m_lastError = rc) ? make_entry_(info) : FileEntry();
}

inline recls_rc_t Search::GetLastError() const
{
    return m_lastError;
}

inline recls_uint32_t Search::GetNumOutstandingDetails() const
{
    recls_uint32_t  cDetails;
    recls_rc_t      rc  =   Recls_OutstandingDetails(m_hSrch, &cDetails);

    if(RECLS_FAILED(m_lastError = rc))
    {
        cDetails = 0;
    }

    return cDetails;
}

inline recls_rc_t &Search::GetLastError_()
{
    return m_lastError;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace cpp */
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !RECLS_INCL_RECLS_CPP_HPP_SEARCH */

/* ////////////////////////////////////////////////////////////////////////// */
