/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls.h
 *
 * Purpose:     Main header file for the recls API.
 *
 * Created:     15th August 2003
 * Updated:     27th December 2005
 *
 * Home:        http://recls.org/
 *
 * Copyright (c) 2003-2005, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/** \file recls.h The root header for the \ref group_recls API */

#ifndef RECLS_INCL_RECLS_H_RECLS
#define RECLS_INCL_RECLS_H_RECLS

/* File version */
#ifndef RECLS_DOCUMENTATION_SKIP_SECTION
# define RECLS_VER_RECLS_H_RECLS_MAJOR      3
# define RECLS_VER_RECLS_H_RECLS_MINOR      0
# define RECLS_VER_RECLS_H_RECLS_REVISION   1
# define RECLS_VER_RECLS_H_RECLS_EDIT       63
#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */

/** \name recls API Version
 *
 * \ingroup group_recls
 */
/** @{ */
/** \def RECLS_VER_MAJOR The major version number of RECLS */

/** \def RECLS_VER_MINOR The minor version number of RECLS */

/** \def RECLS_VER_REVISION The revision version number of RECLS */

/** \def RECLS_VER The current composite version number of RECLS */
/** @} */

/* recls version */
#define RECLS_VER_MAJOR     1
#define RECLS_VER_MINOR     7
#define RECLS_VER_REVISION  1
#define RECLS_VER_1_0_1     0x01000100
#define RECLS_VER_1_1_1     0x01010100
#define RECLS_VER_1_2_1     0x01020100
#define RECLS_VER_1_3_1     0x01030100
#define RECLS_VER_1_4_1     0x01040100
#define RECLS_VER_1_4_2     0x01040200
#define RECLS_VER_1_5_1     0x01050100
#define RECLS_VER_1_5_2     0x01050200
#define RECLS_VER_1_5_3     0x01050300
#define RECLS_VER_1_6_1     0x01060100
#define RECLS_VER_1_6_2     0x01060200
#define RECLS_VER_1_6_3     0x01060300
#define RECLS_VER_1_6_4     0x01060400
#define RECLS_VER_1_6_5     0x01060500
#define RECLS_VER_1_7_1     0x01070100
#define RECLS_VER           RECLS_VER_1_7_1

/* /////////////////////////////////////////////////////////////////////////////
 * Strictness
 */

#ifndef RECLS_NO_STRICT
# define RECLS_STRICT
#endif /* !RECLS_NO_STRICT */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

/* Detects C & C++ things, such as namespace support */
#include <recls/internal/language.h>
/* Includes platform-specific headers */
#include <recls/internal/platform.h>
/* Includes stddef.h / cstddef, and defines the recls types: recls_s/uint8/16/32/64_t */
#include <recls/internal/compiler.h>
/* Defines recls_filesize_t, recls_time_t */
#include <recls/internal/platform_types.h>

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{
#endif /* !RECLS_NO_NAMESPACE */

#if !defined(RECLS_NO_NAMESPACE)
#  define RECLS_QUAL(x)                 ::recls::x
#else
# define RECLS_QUAL(x)                  x
#endif /* !RECLS_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Constants and definitions
 */

/*** Defines the number of characters in a system root. */
#if defined(RECLS_PLATFORM_IS_WIN32)
# define RECLS_ROOT_NAME_LEN            (3)
#elif defined(RECLS_PLATFORM_IS_UNIX)
# define RECLS_ROOT_NAME_LEN            (1)
#elif defined(RECLS_OVERRIDE_PLATFORM)
  /* Assume that the appropriate inclusions are made */
#else
# error Platform not recognised
#endif /* platform */

/* /////////////////////////////////////////////////////////////////////////////
 * Pre-processor discrimination
 */

/* /////////////////////////////////////////////////////////////////////////////
 * Function specifications
 */

/*** Defines the recls linkage and calling convention */
#define RECLS_FNDECL(rt)    RECLS_LINKAGE_C rt RECLS_CALLCONV_DEFAULT

/* /////////////////////////////////////////////////////////////////////////////
 * Typedefs
 */

/* recls_rc_t */
#ifdef __cplusplus

/** The type of return codes issued by the API functions */
typedef recls_sint32_t                                              recls_rc_t;

/** General success code */
const recls_rc_t                                                    RECLS_RC_OK(0);
/** General failure code */
const recls_rc_t                                                    RECLS_RC_FAIL(-1);

/** Returns non-zero if the given return code indicates failure */
inline bool RECLS_FAILED(recls_rc_t const &rc)
{
    return rc < 0;
}

/** Returns non-zero if the given return code indicates success */
inline bool RECLS_SUCCEEDED(recls_rc_t const &rc)
{
    return !RECLS_FAILED(rc);
}

#else /* ? __cplusplus */

/** The type of return codes issued by the API functions */
typedef recls_sint32_t                                              recls_rc_t;

/** General success code */
# define RECLS_RC_OK                                                (0)
/** General failure code */
# define RECLS_RC_FAIL                                              (-1)

/** Evaluates to non-zero if the given return code indicates failure */
# define RECLS_FAILED(rc)                                           ((rc) < 0)
/** Evaluates to non-zero if the given return code indicates success */
# define RECLS_SUCCEEDED(rc)                                        (!RECLS_FAILED(rc))

#endif /* __cplusplus */


/* hrecls_t */
#if !defined(RECLS_DOCUMENTATION_SKIP_SECTION) && \
    defined(RECLS_COMPILER_IS_BORLAND) && \
    __BORLANDC__ < 0x0560
struct hrecls_t_
{
    int dummy;
};
#else /* ? compiler */
struct hrecls_t_;
#endif /* compiler */

/** The handle to a recursive search operation */
typedef struct hrecls_t_ const *                                    hrecls_t;


/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

/* Defines result codes */
#include <recls/internal/retcodes.h>

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{
#endif /* !RECLS_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Enumerations
 */

/** \brief Search flags
 * \ingroup group_recls
 *
 * These flags moderate the search behaviour of the
 * Recls_Search(), Recls_SearchFeedback(), and Recls_SearchProcess() functions.
 */
enum RECLS_FLAG
{
        RECLS_F_FILES                       =   0x00000001  /*!< Include files in search. Included by default if none specified */
    ,   RECLS_F_DIRECTORIES                 =   0x00000002  /*!< Include directories in search */
    ,   RECLS_F_LINKS                       =   0x00000004  /*!< Include links in search. Ignored in Win32 */
    ,   RECLS_F_DEVICES                     =   0x00000008  /*!< Include devices in search. Not currently supported */
    ,   RECLS_F_TYPEMASK                    =   0x00000FFF
    ,   RECLS_F_DIR_PROGRESS                =   0x00001000  /*!< Reports each traversed directory to the callback function supplied to Recls_SearchFeedback() */
    ,   RECLS_F_RECURSIVE                   =   0x00010000  /*!< Searches given directory and all sub-directories */
    ,   RECLS_F_NO_FOLLOW_LINKS             =   0x00020000  /*!< Does not expand links */
    ,   RECLS_F_DIRECTORY_PARTS             =   0x00040000  /*!< Fills out the directory parts. Supported from version 1.1.1 onwards */
    ,   RECLS_F_DETAILS_LATER               =   0x00080000  /*!< Does not fill out anything other than the path. When specified to Recls_Stat() (and no types (RECLS_F_TYPEMASK) are specified, this will allow some details for a non-existant path to be elicited. */
    ,   RECLS_F_PASSIVE_FTP                 =   0x00100000  /*!< Passive mode in FTP. Supported from version 1.5.1 onwards */
    ,   RECLS_F_MARK_DIRS                   =   0x00200000  /*!< Marks the directories with a trailing slash. */
    ,   RECLS_F_ALLOW_REPARSE_DIRS          =   0x00400000  /*!< Allow Win32 reparse point directories to be examined (which can cause infinite loops). */
#if 0
    ,   RECLS_F_CALC_CHECKSUM               =   0x00800000
#endif /* 0 */
    ,   RECLS_F_CALLBACKS_STDCALL_ON_WIN32  =   0x01000000  /*!< Treats the callbacks as stdcall calling convention on Win32. Default is cdecl. */
/*  ,   RECLS_F_DIR_SIZE_IS_NUM_CHILDREN    =   0x02000000  / *!< This causes the  */

#if !defined(FILES)
    ,   FILES                       =   RECLS_F_FILES
#endif /* !FILES */
#if !defined(DIRECTORIES)
    ,   DIRECTORIES                 =   RECLS_F_DIRECTORIES
#endif /* !DIRECTORIES */
#if !defined(LINKS)
    ,   LINKS                       =   RECLS_F_LINKS
#endif /* !LINKS */
#if !defined(DEVICES)
    ,   DEVICES                     =   RECLS_F_DEVICES
#endif /* !DEVICES */
#if !defined(TYPEMASK)
    ,   TYPEMASK                    =   RECLS_F_TYPEMASK
#endif /* !TYPEMASK */
#if !defined(DIR_PROGRESS)
    ,   DIR_PROGRESS                =   RECLS_F_DIR_PROGRESS
#endif /* !DIR_PROGRESS */
#if !defined(RECURSIVE)
    ,   RECURSIVE                   =   RECLS_F_RECURSIVE
#endif /* !RECURSIVE */
#if !defined(NO_FOLLOW_LINKS)
    ,   NO_FOLLOW_LINKS             =   RECLS_F_NO_FOLLOW_LINKS
#endif /* !NO_FOLLOW_LINKS */
#if !defined(DIRECTORY_PARTS)
    ,   DIRECTORY_PARTS             =   RECLS_F_DIRECTORY_PARTS
#endif /* !DIRECTORY_PARTS */
#if !defined(DETAILS_LATER)
    ,   DETAILS_LATER               =   RECLS_F_DETAILS_LATER
#endif /* !DETAILS_LATER */
#if !defined(PASSIVE_FTP)
    ,   PASSIVE_FTP                 =   RECLS_F_PASSIVE_FTP
#endif /* !PASSIVE_FTP */
#if !defined(MARK_DIRS)
    ,   MARK_DIRS                   =   RECLS_F_MARK_DIRS
#endif /* !MARK_DIRS */
#if !defined(ALLOW_REPARSE_DIRS)
    ,   ALLOW_REPARSE_DIRS          =   RECLS_F_ALLOW_REPARSE_DIRS
#endif /* !ALLOW_REPARSE_DIRS */
#if 0
#if !defined(CALC_CHECKSUM)
    ,   CALC_CHECKSUM               =   RECLS_F_CALC_CHECKSUM
#endif /* !CALC_CHECKSUM */
#endif /* 0 */
#if !defined(CALLBACKS_STDCALL_ON_WIN32)
    ,   CALLBACKS_STDCALL_ON_WIN32  =   RECLS_F_CALLBACKS_STDCALL_ON_WIN32
#endif /* !CALLBACKS_STDCALL_ON_WIN32 */
/*  ,   DIR_SIZE_IS_NUM_CHILDREN    =   RECLS_F_DIR_SIZE_IS_NUM_CHILDREN    */
};

/** \brief Selected Roots flags
 * \ingroup group_recls
 *
 * These flags moderate the search behaviour of the Recls_GetSelectedRoots() function.
 *
 * \note The flags are ignored on UNIX, since the only root is '/'
 */
enum RECLS_ROOTS_FLAG
{
        RECLS_F_FIXED_DRIVES        =   0x0001
    ,   RECLS_F_NETWORK_DRIVES      =   0x0002
    ,   RECLS_F_CDROM_DRIVES        =   0x0004
    ,   RECLS_F_REMOVABLE_DRIVES    =   0x0008
    ,   RECLS_F_RAM_DRIVES          =   0x0010
};

#if !defined(__cplusplus) && \
    !defined(RECLS_DOCUMENTATION_SKIP_SECTION)
typedef enum RECLS_FLAG         RECLS_FLAG;
typedef enum RECLS_ROOTS_FLAG   RECLS_ROOTS_FLAG;
#endif /* !__cplusplus && !RECLS_DOCUMENTATION_SKIP_SECTION */


/* /////////////////////////////////////////////////////////////////////////////
 * Typedefs
 */

/** \brief Structure containing information about a file-system root. */

/* #if defined(RECLS_PLATFORM_IS_WIN32)
# define RECLS_ROOT_NAME_LEN            (3)
#elif defined(RECLS_PLATFORM_IS_UNIX)
 */
#ifndef RECLS_COMPILER_IS_CH
typedef struct recls_root
{
    /** The name of the root. */
    recls_char_t    name[1 + RECLS_ROOT_NAME_LEN];
} recls_root_t;
#endif /* !RECLS_COMPILER_IS_CH */

#ifndef RECLS_COMPILER_IS_CH
struct recls_fileinfo_t;
#endif /* !RECLS_COMPILER_IS_CH */

/** Opaque type representing a file-system entry information */
typedef struct recls_fileinfo_t const   *recls_info_t;

/** Opaque type representing a user-defined parameter to the process function */
typedef void                            *recls_process_fn_param_t;

/** User-supplied process function, used by Recls_SearchProcess()
 *
 * \param hEntry entry info structure
 * \param param the parameter passed to Recls_SearchProcess()
 *
 * \return A status to indicate whether to continue or cancel the processing
 * \retval 0 cancel the processing
 * \retval non-0 continue the processing
 */
typedef int (RECLS_CALLCONV_DEFAULT *hrecls_process_fn_t)(  recls_info_t                hEntry
                                                        ,   recls_process_fn_param_t    param);

/** User-supplied process function, used by Recls_SearchFeedback(), which receives the
 * sub-directories traversed
 *
 * \param dir The nul-terminated C-style string containing the sub-directory's full path
 * \param dirLen The number of characters in the \c dir param, not including the nul-terminator
 * \param param the parameter passed to Recls_SearchFeedback()
 *
 * \return A status to indicate whether to continue or cancel the processing
 * \retval 0 cancel the processing
 * \retval non-0 continue the processing
 */
typedef int (RECLS_CALLCONV_DEFAULT *hrecls_progress_fn_t)( recls_char_t const          *dir
                                                        ,   size_t                      dirLen
                                                        ,   recls_process_fn_param_t    param
                                                        ,   void                        *reserved0
                                                        ,   recls_uint32_t              reserved1);

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace typedefs
 */

#if !defined(RECLS_NO_NAMESPACE)
typedef recls_info_t                info_t;
typedef recls_process_fn_param_t    process_fn_param_t;
#endif /* !RECLS_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Functions
 */

/***************************************
 * File system
 */

/** \name File system functions
 *
 * \ingroup group_recls
 */
/** @{ */

#ifndef RECLS_COMPILER_IS_CH
/** Retrieves the file-system roots for the current operating system.
 *
 * It is used as follows:
 *
 * \htmlonly
 * <code>
 * &nbsp;&nbsp;recls_root_t&nbsp;&nbsp;roots[26];<br>
 * &nbsp;&nbsp;size_t&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;cRoots&nbsp;&nbsp;=&nbsp;Recls_GetRoots(roots, 26);<br>
 * &nbsp;&nbsp;size_t&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;i;<br>
 *<br>
 * &nbsp;&nbsp;for(i = 0; i < cRoots; ++i)<br>
 * &nbsp;&nbsp;{<br>
 * &nbsp;&nbsp;&nbsp;&nbsp;puts(roots[i].name);<br>
 * &nbsp;&nbsp;}<br>
 * </code>
 * \endhtmlonly
 *
 * \param roots A pointer to an array of size \c cRoots. If NULL, the function returns the number of available roots.
 * \param cRoots The maximum number of roots to retrive in \c roots.
 *
 * \return The number of roots actually retrieved. If \c roots is NULL, then this is the current number of roots on
 * the current operating system. If \c roots is not NULL, then this is the number of roots written into \c roots, and
 * is not more than \c cRoots.
 */
RECLS_FNDECL(size_t)                Recls_GetRoots(     recls_root_t    *roots
                                                    ,   size_t          cRoots);

/** Like Recls_GetRoots(), but only returns roots of selected disk types
 *
 * \param roots A pointer to an array of size \c cRoots. If NULL, the function returns the number of available roots.
 * \param cRoots The maximum number of roots to retrive in \c roots.
 * \param flags Combination of the RECLS_ROOTS_FLAG enumeration. If 0, implies all available root types.
 *
 * \return The number of roots actually retrieved. If \c roots is NULL, then this is the current number of roots on
 * the current operating system. If \c roots is not NULL, then this is the number of roots written into \c roots, and
 * is not more than \c cRoots.
 */
RECLS_FNDECL(size_t)                Recls_GetSelectedRoots( recls_root_t    *roots
                                                        ,   size_t          cRoots
                                                        ,   recls_uint32_t  flags);
#endif /* !RECLS_COMPILER_IS_CH */

/** Returns the symbol used to separate distinct path names in path name lists.
 *
 * \note On UNIX this is \c "/"; on Win32 it is \c "\".
 *
 * It is used as follows:
 *
 * \htmlonly
 * <code>
 * &nbsp;&nbsp;puts(Recls_GetPathNameSeparator());<br>
 * </code>
 * \endhtmlonly
 */
RECLS_FNDECL(recls_char_t const*)   Recls_GetPathNameSeparator(void);

/** Returns the symbol used to separate the directory parts within paths.
 *
 * \note On UNIX this is \c ":"; on Win32 it is \c ";".
 *
 * It is used as follows:
 *
 * \htmlonly
 * <code>
 * &nbsp;&nbsp;puts(Recls_GetPathSeparator());<br>
 * </code>
 * \endhtmlonly
 */
RECLS_FNDECL(recls_char_t const*)   Recls_GetPathSeparator(void);

/** Returns the wildcard symbol used to represent the "all files" for the current operating system.
 *
 * \note On UNIX this is \c "*"; on Win32 it is \c "*.*".
 *
 * It is used as follows:
 *
 * \htmlonly
 * <code>
 * &nbsp;&nbsp;puts(Recls_GetWildcardsAll());<br>
 * </code>
 * \endhtmlonly
 */
RECLS_FNDECL(recls_char_t const*)   Recls_GetWildcardsAll(void);

/** @} */

/***************************************
 * Search control
 */

/** \name Search control functions
 *
 * \ingroup group_recls
 */
/** @{ */

/** Searches a given directory for matching files of the given pattern
 *
 *
 * \param searchRoot The directory representing the root of the search
 * \param pattern The search pattern, e.g. "*.c"
 * \param flags A combination of 0 or more RECLS_FLAG values
 * \param phSrch Address of the search handle. This is set to NULL on failure
 *
 * \return A status code indicating success/failure
 *
 */
RECLS_FNDECL(recls_rc_t) Recls_Search(          recls_char_t const          *searchRoot
                                    ,           recls_char_t const          *pattern
                                    ,           recls_uint32_t              flags
                                    ,           hrecls_t                    *phSrch);

/** Searches a given directory for matching files of the given pattern, calling back 
 * on the given progress function pointer to inform the caller as each (sub-)directory
 * is traversed.
 *
 * \param searchRoot The directory representing the root of the search
 * \param pattern The search pattern, e.g. "*.c"
 * \param flags A combination of 0 or more RECLS_FLAG values
 * \param phSrch Address of the search handle. This is set to NULL on failure
 *
 * \return A status code indicating success/failure
 */
RECLS_FNDECL(recls_rc_t) Recls_SearchFeedback(  recls_char_t const          *searchRoot
                                    ,           recls_char_t const          *pattern
                                    ,           recls_uint32_t              flags
                                    ,           hrecls_progress_fn_t        pfn
                                    ,           recls_process_fn_param_t    param
                                    ,           hrecls_t                    *phSrch);


/** Searches a given directory for matching files of the given pattern, and processes them according to the given process function
 *
 * \param searchRoot The directory representing the root of the search
 * \param pattern The search pattern, e.g. "*.c"
 * \param flags A combination of 0 or more RECLS_FLAG values
 * \param pfn The processing function
 * \param param A caller-supplied parameter that is passed through to \c pfn on each invocation. The function can cancel the enumeration by returning 0
 *
 * \return A status code indicating success/failure
 *
 * \note Available from version 1.1 of the <b>recls</b> API
 */
RECLS_FNDECL(recls_rc_t) Recls_SearchProcess(   recls_char_t const          *searchRoot
                                            ,   recls_char_t const          *pattern
                                            ,   recls_uint32_t              flags
                                            ,   hrecls_process_fn_t         pfn
                                            ,   recls_process_fn_param_t    param);

/** Closes the given search
 *
 * \param hSrch Handle of the search to close. May not be NULL.
 */
RECLS_FNDECL(void) Recls_SearchClose(           hrecls_t                    hSrch);


/** Retrieves the information for the given path
 *
 * \param path Absolute or relative path of the file system entry whose information is to be retrieved.
 * \param flags Flags to moderate the search. Only the type flags (flags & RECLS_F_TYPEMASK) and RECLS_F_DIRECTORY_PARTS are recognised
 * \param phEntry Pointer to receive entry info structure.
 *
 * \return Status code
 * \retval RECLS_RC_OK Position was advanced; search handle can be queried for details
 * \retval RECLS_RC_NO_MORE_DATA There was no file system entry that matched the \c path and \c flags criteria
 * \retval Any other status code indicates an error
 */
RECLS_FNDECL(recls_rc_t) Recls_Stat(            recls_char_t const          *path
                                            ,   recls_uint32_t              flags
                                            ,   recls_info_t                *phEntry);

/** @} */

/***************************************
 * Search enumeration
 */

/** \name Search enumeration functions
 *
 * \ingroup group_recls
 */
/** @{ */

/** Advances the search one position
 *
 * \param hSrch Handle of the search to close. May not be NULL.
 *
 * \return Status code
 * \retval RECLS_RC_OK Position was advanced; search handle can be queried for details
 * \retval RECLS_RC_NO_MORE_DATA There are no more items in the search
 * \retval Any other status code indicates an error
 */
RECLS_FNDECL(recls_rc_t) Recls_GetNext(         hrecls_t                    hSrch);

/** Retrieves the information for the current search position
 *
 * \param hSrch Handle of the search to close. May not be NULL.
 * \param phEntry Pointer to receive entry info structure.
 *
 * \return Status code
 * \retval RECLS_RC_OK Position was advanced; search handle can be queried for details
 * \retval RECLS_RC_NO_MORE_DATA There are no more items in the search
 * \retval Any other status code indicates an error
 */
RECLS_FNDECL(recls_rc_t) Recls_GetDetails(      hrecls_t                    hSrch
                                            ,   recls_info_t                *phEntry);

/** Advances the search one position, and retrieves the information for the new position
 *
 * \param hSrch Handle of the search to close. May not be NULL.
 * \param phEntry Pointer to receive entry info structure.
 *
 * \return Status code
 * \retval RECLS_RC_OK Position was advanced; search handle can be queried for details
 * \retval Any other status code indicates an error
 */
RECLS_FNDECL(recls_rc_t) Recls_GetNextDetails(  hrecls_t                    hSrch
                                            ,   recls_info_t                *phEntry);

/** @} */

/***************************************
 * File entry information
 */

/** \name File entry information functions
 *
 * \ingroup group_recls
 */
/** @{ */

/** Releases the resources associated with an entry info structure.
 *
 * \param hEntry The info entry structure handle. Must not be NULL.
 */
RECLS_FNDECL(void) Recls_CloseDetails(          recls_info_t                hEntry);

/** Copies an entry info structure.
 *
 * \param hEntry The info entry structure.
 * \param phEntry Address to receive a copy of the info entry structure. May not be NULL.
 *
 * \return Status code
 * \retval RECLS_RC_OK Entry was generated.
 * \retval Any other status code indicates an error
 */
RECLS_FNDECL(recls_rc_t) Recls_CopyDetails(     recls_info_t                hEntry
                                            ,   recls_info_t                *phEntry);

/** Reports on the number of outstanding (i.e. in client code) file entry info structures
 *
 * \param hSrch Handle of the search to close. May not be NULL.
 * \param count Pointer to an integer variable to receive the result.
 *
 * \return Status code
 * \retval RECLS_RC_OK Information was retrieved.
 * \retval Any other status code indicates an error
 */
RECLS_FNDECL(recls_rc_t) Recls_OutstandingDetails(hrecls_t                  hSrch
                                            ,   recls_uint32_t              *count);

/** @} */

/***************************************
 * Error handling
 */

/** \name Error handling functions
 *
 * \ingroup group_recls
 */
/** @{ */

/** Returns the last error code associated with the given search handle
 *
 * \param hSrch Handle of the search to close. May not be NULL.
 *
 * \return The last error code for the search handle
 */
RECLS_FNDECL(recls_rc_t) Recls_GetLastError(    hrecls_t                    hSrch);

/** Gets the error string representing the given error
 *
 * \param rc The error code
 * \param buffer Pointer to character buffer in which to write the error. If NULL, the function returns the number of characters required.
 * \param cchBuffer Number of character spaces in \c buffer. Ignored if \c buffer is NULL.
 *
 * \return The number of characters written to the buffer, or required for the error.
 */
RECLS_FNDECL(size_t) Recls_GetErrorString(      /* [in] */ recls_rc_t       rc
                                    ,           /* [in] */ recls_char_t     *buffer
                                    ,           /* [in] */ size_t           cchBuffer);

/** Gets the error string representing the current error associated with the given search handle
 *
 * \param hSrch Handle of the search to close. May not be NULL.
 * \param buffer Pointer to character buffer in which to write the error. If NULL, the function returns the number of characters required.
 * \param cchBuffer Number of character spaces in \c buffer. Ignored if \c buffer is NULL.
 *
 * \return The number of characters written to the buffer, or required for, the error.
 */
RECLS_FNDECL(size_t) Recls_GetLastErrorString(  /* [in] */ hrecls_t         hSrch
                                        ,       /* [in] */ recls_char_t     *buffer
                                        ,       /* [in] */ size_t           cchBuffer);

/** @} */

/***************************************
 * Property elicitation
 */

/** \name Property elicitation functions
 *
 * \ingroup group_recls
 */
/** @{ */

/** Retrieves the full path of the given entry recls_fileinfo_t
 *
 * \param hEntry The entry recls_fileinfo_t. Cannot be NULL
 * \param buffer Pointer to character buffer in which to write the path. If NULL, the function returns the number of characters required.
 * \param cchBuffer Number of character spaces in \c buffer. Ignored if \c buffer is NULL.
 *
 * \return The number of characters written to the buffer, or required for, the path.
 */
RECLS_FNDECL(size_t) Recls_GetPathProperty(     recls_info_t                hEntry
                                        ,       recls_char_t                *buffer
                                        ,       size_t                      cchBuffer);

/** Retrieves the path of the given entry recls_fileinfo_t, relative to the search directory
 *
 * \param hEntry The entry recls_fileinfo_t. Cannot be NULL
 * \param buffer Pointer to character buffer in which to write the path. If NULL, the function returns the number of characters required.
 * \param cchBuffer Number of character spaces in \c buffer. Ignored if \c buffer is NULL.
 *
 * \return The number of characters written to the buffer, or required for, the path.
 */
RECLS_FNDECL(size_t) Recls_GetSearchRelativePathProperty(   recls_info_t    hEntry
                                                    ,       recls_char_t    *buffer
                                                    ,       size_t          cchBuffer);

/** Retrieves the directory of the given entry recls_fileinfo_t
 *
 * \param hEntry The entry recls_fileinfo_t. Cannot be NULL
 * \param buffer Pointer to character buffer in which to write the directory. If NULL, the function returns the number of characters required.
 * \param cchBuffer Number of character spaces in \c buffer. Ignored if \c buffer is NULL.
 *
 * \return The number of characters written to the buffer, or required for, the directory.
 */
RECLS_FNDECL(size_t) Recls_GetDirectoryProperty(recls_info_t                hEntry
                                            ,   recls_char_t                *buffer
                                            ,   size_t                      cchBuffer);

/** Retrieves the directory (and drive) of the given entry recls_fileinfo_t
 *
 * \param hEntry The entry recls_fileinfo_t. Cannot be NULL
 * \param buffer Pointer to character buffer in which to write the directory. If NULL, the function returns the number of characters required.
 * \param cchBuffer Number of character spaces in \c buffer. Ignored if \c buffer is NULL.
 *
 * \return The number of characters written to the buffer, or required for, the directory.
 *
 * \note On systems that do not have a drive, this function behaves identically to Recls_GetDirectoryProperty()
 */
RECLS_FNDECL(size_t) Recls_GetDirectoryPathProperty(    recls_info_t        hEntry
                                                    ,   recls_char_t        *buffer
                                                    ,   size_t              cchBuffer);

/** Retrieves the directory (and drive) of the given entry recls_fileinfo_t that was used in the search resulting in the given entry
 *
 * \param hEntry The entry recls_fileinfo_t. Cannot be NULL
 * \param buffer Pointer to character buffer in which to write the directory. If NULL, the function returns the number of characters required.
 * \param cchBuffer Number of character spaces in \c buffer. Ignored if \c buffer is NULL.
 *
 * \return The number of characters written to the buffer, or required for, the directory.
 */
RECLS_FNDECL(size_t) Recls_GetSearchDirectoryProperty(  recls_info_t        hEntry
                                                    ,   recls_char_t        *buffer
                                                    ,   size_t              cchBuffer);

/** Retrieves the UNC drive property of the given entry recls_fileinfo_t
 *
 * \param hEntry The entry recls_fileinfo_t. Cannot be NULL
 * \param buffer Pointer to character buffer in which to write the directory. If NULL, the function returns the number of characters required.
 * \param cchBuffer Number of character spaces in \c buffer. Ignored if \c buffer is NULL.
 *
 * \return The number of characters written to the buffer, or required for, the directory.
 *
 * \note On operating systems that do not support UNC, this function always returns 0, and writes the empty string into buffer (when non-NULL)
 */
RECLS_FNDECL(size_t) Recls_GetUNCDriveProperty(         recls_info_t        hEntry
                                                    ,   recls_char_t        *buffer
                                                    ,   size_t              cchBuffer);

/** Retrieves the file (filename + extension) of the given entry recls_fileinfo_t
 *
 * \param hEntry The entry recls_fileinfo_t. Cannot be NULL
 * \param buffer Pointer to character buffer in which to write the file. If NULL, the function returns the number of characters required.
 * \param cchBuffer Number of character spaces in \c buffer. Ignored if \c buffer is NULL.
 *
 * \return The number of characters written to the buffer, or required for, the file.
 */
RECLS_FNDECL(size_t) Recls_GetFileProperty(     recls_info_t                hEntry
                                        ,       recls_char_t                *buffer
                                        ,       size_t                      cchBuffer);

/** Retrieves the short version of the file of the given entry recls_fileinfo_t
 *
 * \param hEntry The entry recls_fileinfo_t. Cannot be NULL
 * \param buffer Pointer to character buffer in which to write the file. If NULL, the function returns the number of characters required.
 * \param cchBuffer Number of character spaces in \c buffer. Ignored if \c buffer is NULL.
 *
 * \return The number of characters written to the buffer, or required for, the file.
 *
 * \note On systems where there is no concept of a short name, this function behaves exactly as Recls_GetFileProperty()
 */
RECLS_FNDECL(size_t) Recls_GetShortFileProperty(recls_info_t                hEntry
                                            ,   recls_char_t                *buffer
                                            ,   size_t                      cchBuffer);

/** Retrieves the filename (not including extension, if any) of the given entry recls_fileinfo_t
 *
 * \param hEntry The entry recls_fileinfo_t. Cannot be NULL
 * \param buffer Pointer to character buffer in which to write the filename. If NULL, the function returns the number of characters required.
 * \param cchBuffer Number of character spaces in \c buffer. Ignored if \c buffer is NULL.
 *
 * \return The number of characters written to the buffer, or required for, the filename.
 */
RECLS_FNDECL(size_t) Recls_GetFileNameProperty( recls_info_t                hEntry
                                            ,   recls_char_t                *buffer
                                            ,   size_t                      cchBuffer);

/** Retrieves the file extension of the given entry recls_fileinfo_t
 *
 * \param hEntry The entry recls_fileinfo_t. Cannot be NULL
 * \param buffer Pointer to character buffer in which to write the extension. If NULL, the function returns the number of characters required.
 * \param cchBuffer Number of character spaces in \c buffer. Ignored if \c buffer is NULL.
 *
 * \return The number of characters written to the buffer, or required for, the extension.
 */
RECLS_FNDECL(size_t) Recls_GetFileExtProperty(  recls_info_t                hEntry
                                            ,   recls_char_t                *buffer
                                            ,   size_t                      cchBuffer);

/** Retrieves a directory part of the given entry recls_fileinfo_t
 *
 * \param hEntry The entry recls_fileinfo_t. Cannot be NULL
 * \param part The part requested. If -1, then the function returns the number of parts
 * \param buffer Pointer to character buffer in which to write the extension. If NULL, the function returns the number of characters required. Ignored if part is -1.
 * \param cchBuffer Number of character spaces in \c buffer. Ignored if \c buffer is NULL or part is -1.
 *
 * \return If \c part is -1, returns the number of parts. Otherwise, The number of characters written to the buffer, or required for, the extension.
 *
 * \note The behaviour is undefined if part is outside the range of parts.
 */
RECLS_FNDECL(size_t) Recls_GetDirectoryPartProperty(recls_info_t            hEntry
                                            ,   int                         part
                                            ,   recls_char_t                *buffer
                                            ,   size_t                      cchBuffer);


/** Returns non-zero if the file entry is read-only.
 *
 * \param hEntry The file entry info structure to test. May not be NULL
 * \retval true file entry is read-only
 * \retval false file entry is not read-only
 *
 * \note There is no error return
 */
RECLS_FNDECL(recls_bool_t) Recls_IsFileReadOnly(recls_info_t                hEntry);

/** Returns non-zero if the file entry represents a directory.
 *
 * \param hEntry The file entry info structure to test. May not be NULL
 * \retval true file entry is a directory
 * \retval false file entry is not directory
 *
 * \note There is no error return
 */
RECLS_FNDECL(recls_bool_t) Recls_IsFileDirectory(recls_info_t               hEntry);

/** Returns non-zero if the file entry represents a link.
 *
 * \param hEntry The file entry info structure to test. May not be NULL
 * \retval true file entry is a link
 * \retval false file entry is not link
 *
 * \note There is no error return
 */
RECLS_FNDECL(recls_bool_t) Recls_IsFileLink(    recls_info_t                hEntry);

/** Returns non-zero if the file entry's path is UNC.
 *
 * \param hEntry The file entry info structure to test. May not be NULL
 * \retval true file entry's path is UNC
 * \retval false file entry's path is UNC
 *
 * \note There is no error return
 */
RECLS_FNDECL(recls_bool_t) Recls_IsFileUNC(     recls_info_t                hEntry);

/** Acquires the size of the file entry.
 *
 * \param hEntry The file entry info structure to test. May not be NULL
 * \param size Pointer to the location in which to store the size
 *
 * \note There is no error return. File system entries that do not have a meaningful size will be given a notional size of 0.
 */
RECLS_FNDECL(void)         Recls_GetSizeProperty(   recls_info_t            hEntry
                                                ,   recls_filesize_t        *size);

/** Returns the time the file was created */
RECLS_FNDECL(recls_time_t) Recls_GetCreationTime(recls_info_t               hEntry);

/** Returns the time the file was last modified */
RECLS_FNDECL(recls_time_t) Recls_GetModificationTime(recls_info_t           hEntry);

/** Returns the time the file was last accessed */
RECLS_FNDECL(recls_time_t) Recls_GetLastAccessTime(recls_info_t             hEntry);

/** Returns the time the file status was last changed */
RECLS_FNDECL(recls_time_t) Recls_GetLastStatusChangeTime(recls_info_t       hEntry);

#if 0
/** Returns the checksum value of the file.
 *
 * \note This will return 0 if the entry is not a file, or its contents could
 * not be read, or RECLS_F_CALC_CHECKSUM was not specified.
 */
RECLS_FNDECL(recls_uint32_t) Recls_GetChecksum(         recls_info_t        hEntry);
#endif /* 0 */

/** @} */

/***************************************
 * Extended API functions
 */

/** \name Extended API functions
 *
 * \ingroup group_recls
 */
/** @{ */

/* \brief Determines whether the given directory is empty
 *
 * \param dir The directory to assess
 * \note New with recls version 1.4.2
 */
RECLS_FNDECL(recls_bool_t) Recls_IsDirectoryEmpty(recls_char_t const        *dir);

RECLS_FNDECL(recls_bool_t) Recls_IsDirectoryEntryEmpty(recls_info_t         hEntry);

RECLS_FNDECL(recls_filesize_t) Recls_CalcDirectorySize(recls_char_t const   *dir);

RECLS_FNDECL(recls_filesize_t) Recls_CalcDirectoryEntrySize(recls_info_t    hEntry);

/** @} */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

#ifdef __cplusplus

# ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
# endif /* !_STLSOFT_NO_NAMESPACE */

# if defined(RECLS_NO_NAMESPACE)
inline recls_char_t const *c_str_ptr(recls_root_t const &root)
# else /* ? RECLS_NO_NAMESPACE */
inline ::recls::recls_char_t const *c_str_ptr(::recls::recls_root_t const &root)
# endif /* RECLS_NO_NAMESPACE */
{
    return root.name;
}

# ifndef _STLSOFT_NO_NAMESPACE
} /* namespace stlsoft */
# endif /* !_STLSOFT_NO_NAMESPACE */

#endif /* __cplusplus */

/* /////////////////////////////////////////////////////////////////////////////
 * Platform-specific includes
 */

/*** \def RECLS_PLATFORM_API_WIN32 Defined if Win32 platform-specific extensions are in use */
/*** \def RECLS_PLATFORM_API_UNIX Defined if UNIX platform-specific extensions are in use */

#ifdef RECLS_PLATFORM_API_WIN32
# undef RECLS_PLATFORM_API_WIN32
#endif /* RECLS_PLATFORM_API_WIN32 */

#ifdef RECLS_PLATFORM_API_UNIX
# undef RECLS_PLATFORM_API_UNIX
#endif /* RECLS_PLATFORM_API_UNIX */

#if !defined(RECLS_PURE_API)
# if defined(RECLS_PLATFORM_IS_WIN32)
#  include <recls/win32.h>
#  define RECLS_PLATFORM_API_WIN32
# elif defined(RECLS_PLATFORM_IS_UNIX)
#  include <recls/unix.h>
#  define RECLS_PLATFORM_API_UNIX
# else
#  error Platform not recognised
# endif /* platform */
#endif /* !RECLS_PURE_API */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !RECLS_INCL_RECLS_H_RECLS */

/* ////////////////////////////////////////////////////////////////////////// */
