/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls/stl/traits.hpp
 *
 * Purpose:     Contains character-base traits for the recls::stl mapping.
 *
 * Created:     25th March 2004
 * Updated:     27th December 2005
 *
 * Home:        http://recls.org/
 *
 * Copyright (c) 2004-2005, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef RECLS_INCL_RECLSTL_STL_HPP_TRAITS
#define RECLS_INCL_RECLSTL_STL_HPP_TRAITS

#ifndef RECLS_DOCUMENTATION_SKIP_SECTION
# define RECLS_VER_RECLSTL_STL_HPP_TRAITS_MAJOR     4
# define RECLS_VER_RECLSTL_STL_HPP_TRAITS_MINOR     0
# define RECLS_VER_RECLSTL_STL_HPP_TRAITS_REVISION  1
# define RECLS_VER_RECLSTL_STL_HPP_TRAITS_EDIT      14
#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */

/** \file recls/stl/traits.hpp Definition of the \link recls::stl::reclstl_traits reclstl_traits\endlink 
 * traits class 
 * (\link recls::stl::basic_search_sequence_const_iterator basic_search_sequence_const_iterator\endlink and 
 * \link recls::stl::basic_search_sequence_value_type basic_search_sequence_value_type\endlink), representing the
 * \ref group_recls_stl "recls STL mapping" */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef RECLS_INCL_RECLS_STL_HPP_RECLS
# include <recls/stl/recls.hpp>
#endif /* !RECLS_INCL_RECLS_STL_HPP_RECLS */
#ifndef RECLS_INCL_H_RECLS_ASSERT
# include <recls/assert.h>
#endif /* !RECLS_INCL_H_RECLS_ASSERT */
#include <stlsoft/proxy_sequence.hpp>

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{

namespace stl
{
#endif /* !RECLS_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

/// \brief A traits class to provide character-encoding specific mappings from the \ref group_recls "recls C API" to the \ref group_recls_stl "recls.stl" mapping"
/// \ingroup group_recls_stl
/// 
template <typename C>
struct reclstl_traits
{
public:
    /// The <code>char</code> type
    ///
    /// \note In this template, <code>void</code> is a placeholder. <code>recls_char_a_t</code> and <code>recls_char_w_t</code> specialisations
    /// will use their respective character types
    typedef void            char_type;
    /// The entry type
    ///
    /// \note In this template, <code>void*</code> is a placeholder. <code>recls_char_a_t</code> and <code>recls_char_w_t</code> specialisations
    /// will use their respective \c recls_fileinfo_t types
    typedef void            *entry_type;

public:
#if defined(RECLS_API_FTP)
    /// traits-form of \link recls::Recls_SearchFtp Recls_SearchFtp\endlink
    static recls_rc_t       SearchFtp(char_type const *host, char_type const *username, char_type const *password, char_type const *searchRoot, char_type const *pattern, recls_uint32_t flags, hrecls_t *phSrch);
#endif /* RECLS_API_FTP */

    /// traits-form of \link recls::Recls_Search Recls_Search\endlink
    static recls_rc_t       Search(char_type const *searchRoot, char_type const *pattern, recls_uint32_t flags, hrecls_t *phSrch);

    /// traits-form of \link recls::Recls_GetDetails Recls_GetDetails\endlink
    static recls_rc_t       GetDetails(hrecls_t hSrch, entry_type *pinfo);
    /// traits-form of \link recls::Recls_GetNextDetails Recls_GetNextDetails\endlink
    static recls_rc_t       GetNextDetails(hrecls_t hSrch, entry_type *pinfo);

    /// traits-form of \link recls::Recls_CloseDetails Recls_CloseDetails\endlink
    static void             CloseDetails(entry_type fileInfo);
    /// traits-form of \link recls::Recls_CopyDetails Recls_CopyDetails\endlink
    static entry_type       CopyDetails(entry_type fileInfo);

    /// traits-form of <code>strcpy</code>
    static char_type        *str_copy(char_type *, char_type const *);
};

#ifndef RECLS_DOCUMENTATION_SKIP_SECTION

# ifdef RECLS_CHAR_TYPE_IS_CHAR

template <>
struct reclstl_traits<recls_char_a_t>
{
public:
    typedef reclstl_traits<recls_char_a_t>      traits_type;
    typedef recls_char_a_t                      char_type;
    typedef recls_info_t                        entry_type;
    typedef recls_strptrs_t                     strptrs_type;
    typedef recls_strptrsptrs_t                 strptrsptrs_type;

    /// The directory parts sequence type
    typedef stlsoft::proxy_sequence<    const strptrs_type
                                    ,   string_t
                                    ,   traits_type // Note, this type also serves as the traits type for the STLSoft proxy_sequence
                                    >           directory_parts_type;

    /// Constructs a value of type \c string_t from an instance of the pointers structure.
    ///
    /// \note This is required by the STLSoft proxy_sequence template.
    static string_t make_value(strptrs_type const &ptrs)
    {
        return string_t(ptrs.begin, ptrs.end);
    }

public:
#if defined(RECLS_API_FTP)
    static recls_rc_t SearchFtp(char_type const *host, char_type const *username, char_type const *password, char_type const *searchRoot, char_type const *pattern, recls_uint32_t flags, hrecls_t *phSrch)
    {
        return Recls_SearchFtp(host, username, password, searchRoot, pattern, flags, phSrch);
    }
#endif /* RECLS_API_FTP */

    static recls_rc_t Search(char_type const *searchRoot, char_type const *pattern, recls_uint32_t flags, hrecls_t *phSrch)
    {
        return Recls_Search(searchRoot, pattern, flags, phSrch);
    }

    static recls_rc_t GetDetails(hrecls_t hSrch, entry_type *pinfo)
    {
        return Recls_GetDetails(hSrch, pinfo);
    }

    static recls_rc_t GetNextDetails(hrecls_t hSrch, entry_type *pinfo)
    {
        return Recls_GetNextDetails(hSrch, pinfo);
    }

    static void CloseDetails(entry_type fileInfo)
    {
        Recls_CloseDetails(fileInfo);
    }
    static entry_type CopyDetails(entry_type fileInfo)
    {
        entry_type  infoCopy;

        return RECLS_SUCCEEDED(Recls_CopyDetails(fileInfo, &infoCopy)) ? infoCopy : static_cast<entry_type>(NULL);
    }

    static char_type *str_copy(char_type *dest, char_type const *src)
    {
        return strcpy(dest, src);
    }
};

# elif defined(RECLS_CHAR_TYPE_IS_WCHAR)

template <>
struct reclstl_traits<recls_char_w_t>
{
public:
    typedef reclstl_traits<recls_char_w_t>      traits_type;
    typedef recls_char_w_t                      char_type;
    typedef recls_info_t                        entry_type;
    typedef recls_strptrs_t                     strptrs_type;
    typedef recls_strptrsptrs_t                 strptrsptrs_type;

    /// The directory parts sequence type
    typedef stlsoft::proxy_sequence<    const strptrs_type
                                    ,   string_t
                                    ,   traits_type // Note, this type also serves as the traits type for the STLSoft proxy_sequence
                                    >           directory_parts_type;

    /// Constructs a value of type \c string_t from an instance of the pointers structure.
    ///
    /// \note This is required by the STLSoft proxy_sequence template.
    static string_t make_value(strptrs_type const &ptrs)
    {
        return string_t(ptrs.begin, ptrs.end);
    }

public:
#if defined(RECLS_API_FTP)
    static hrecls_t SearchFtp(char_type const *host, char_type const *username, char_type const *password, char_type const *searchRoot, char_type const *pattern, recls_uint32_t flags)
    {
        hrecls_t    hSrch;
        recls_rc_t  rc = Recls_SearchFtp(host, username, password, searchRoot, pattern, flags, &hSrch);

        return RECLS_SUCCEEDED(rc) ? hSrch : static_cast<hrecls_t>(NULL);
    }
#endif /* RECLS_API_FTP */

    static hrecls_t Search(char_type const *searchRoot, char_type const *pattern, recls_uint32_t flags)
    {
        hrecls_t    hSrch;
        recls_rc_t  rc = Recls_Search(searchRoot, pattern, flags, &hSrch);

        return RECLS_SUCCEEDED(rc) ? hSrch : static_cast<hrecls_t>(NULL);
    }

    static recls_rc_t GetDetails(hrecls_t hSrch, entry_type *pinfo)
    {
        return Recls_GetDetails(hSrch, pinfo);
    }

    static recls_rc_t GetNextDetails(hrecls_t hSrch, entry_type *pinfo)
    {
        return Recls_GetNextDetails(hSrch, pinfo);
    }

    static void CloseDetails(entry_type fileInfo)
    {
        Recls_CloseDetails(fileInfo);
    }
    static entry_type CopyDetails(entry_type fileInfo)
    {
        entry_type  infoCopy;

        return RECLS_SUCCEEDED(Recls_CopyDetails(fileInfo, &infoCopy)) ? infoCopy : static_cast<entry_type>(NULL);
    }

    static char_type *str_copy(char_type *dest, char_type const *src)
    {
        return wcscpy(dest, src);
    }
};

# else /* character type */

#  error Neither RECLS_CHAR_TYPE_IS_WCHAR nor RECLS_CHAR_TYPE_IS_CHAR are defined

# endif /* character type */


#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */

/* ////////////////////////////////////////////////////////////////////////// */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace stl */
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* RECLS_INCL_RECLSTL_STL_HPP_TRAITS */

/* ////////////////////////////////////////////////////////////////////////// */
