////////////////////////////////////////////////////////////////////////////////
// File:	api.cs
//
// Created: 24th September 2003
// Updated: 3rd December 2005
//
////////////////////////////////////////////////////////////////////////////////

namespace recls
{
	using System;
	using System.Runtime.InteropServices;
	using System.Text;
	using hrecls_t			=	System.IntPtr;
	using recls_info_t		=	System.IntPtr;
	using recls_filesize_t	=	System.UInt64;

	/// <summary>
	/// A callback method that receives information about the current progress of a file-system search.
	/// </summary>
	public delegate int ProgressFnCallback(string dir, uint dirLen, int param, int reserved0, uint reserved1);

	/// <summary>
	/// recls search flags
	/// </summary>
	/// <remarks>
	/// <p>
	/// These flags alter the behaviour of a given search. Some of the flags affect each other, such
	/// as <see cref="RECLS_F_FILES"/>, <see cref="RECLS_F_DIRECTORIES"/>, <see cref="RECLS_F_LINKS"/> and
	/// <see cref="RECLS_F_DEVICES"/>, which together encompass the possible types of file entries retrieved
	/// by a search.
	/// </p>
	/// <p>
	/// <b>Note:</b> if none of these flags are specified, <see cref="RECLS_F_FILES"/> will be assumed
	/// </p>
	/// <p>
	/// <b>Note:</b> <see cref="RECLS_F_LINKS"/> and <see cref="RECLS_F_DEVICES"/> are not currently supported
	/// </p>
	/// <p>
	/// The other flags moderate how the search is conducted:
	/// </p>
	/// <p>
	/// <see cref="RECLS_F_RECURSIVE"/> causes the search to be conducted on the given directory, and all its sub-directories
	/// </p>
	/// <p>
	/// <see cref="RECLS_F_NO_FOLLOW_LINKS"/> is not currently supported
	/// </p>
	/// <p>
	/// <see cref="RECLS_F_DIRECTORY_PARTS"/> causes the components of the directory for each entry to be calculated, which are
	/// then available as the <see cref="FileEntry.DirectoryParts"/> property
	/// </p>
	/// <p>
	/// <see cref="RECLS_F_DETAILS_LATER"/> is not currently supported
	/// </p>
	/// </remarks>
	[Flags]
	public enum RECLS_FLAG
	{
			/// <summary>
			/// Include files in search. Included by default if none specified
			/// </summary>
			FILES						=   0x00000001
		,	/// <summary> Obsolete. Equal to FILES </summary>
			RECLS_F_FILES				=	FILES

		,	/// <summary>
			/// Include directories in search
			/// </summary>
			DIRECTORIES					=   0x00000002
		,	/// <summary> Obsolete. Equal to FILES </summary>
			RECLS_F_DIRECTORIES			=	DIRECTORIES

		,	/// <summary>
			/// Include links in search
			/// </summary>
			LINKS						=   0x00000004
		,	/// <summary> Obsolete. Equal to FILES </summary>
			RECLS_F_LINKS				=   LINKS

		,	/// <summary>
			/// Include devices in search
			/// </summary>
			DEVICES						=   0x00000008
		,	/// <summary> Obsolete. Equal to DEVICES </summary>
			RECLS_F_DEVICES             =   DEVICES

		,	/// <summary>
			/// Searches given directory and all sub-directories
			/// </summary>
			RECURSIVE					=   0x00010000
		,	/// <summary> Obsolete. Equal to DEVICES </summary>
			RECLS_F_RECURSIVE           =   RECURSIVE

		,	/// <summary>
			/// Does not expand links
			/// </summary>
			NO_FOLLOW_LINKS				=   0x00020000
		,	/// <summary> Obsolete. Equal to NO_FOLLOW_LINKS </summary>
			RECLS_F_NO_FOLLOW_LINKS     =   NO_FOLLOW_LINKS

		,	/// <summary>
			/// Fills out the directory parts
			/// </summary>
			DIRECTORY_PARTS				=   0x00040000
		,	/// <summary> Obsolete. Equal to DIRECTORY_PARTS </summary>
			RECLS_F_DIRECTORY_PARTS     =   DIRECTORY_PARTS

		,	/// <summary>
			/// Does not fill out anything other than the path
			/// </summary>
			DETAILS_LATER				=   0x00080000
	   ,	/// <summary> Obsolete. Equal to DIRECTORY_PARTS </summary>
			RECLS_F_DETAILS_LATER		=   DETAILS_LATER

		,	/// <summary>
			/// Passive mode in FTP. Supported from version 1.5.1 onwards
			/// </summary>
			PASSIVE_FTP					=   0x00100000
		,	/// <summary> Obsolete. Equal to PASSIVE_FTP </summary>
			RECLS_F_PASSIVE_FTP         =   PASSIVE_FTP

		,	/// <summary>
			/// Marks the directories with a trailing slash
			/// </summary>
			MARK_DIRS					=   0x00200000
		,	/// <summary> Obsolete. Equal to MARK_DIRS </summary>
			RECLS_F_MARK_DIRS           =   MARK_DIRS

		,	/// <summary>
			/// Allow Win32 reparse point directories to be examined (which can cause infinite loops)
			/// </summary>
			ALLOW_REPARSE_DIRS			=   0x00400000
		,	/// <summary> Obsolete. Equal to ALLOW_REPARSE_DIRS </summary>
			RECLS_F_ALLOW_REPARSE_DIRS  =   ALLOW_REPARSE_DIRS
	};

	/// <summary>
	/// The return codes emitted by the recls API, and carried by
	/// ReclsException instances.
	/// </summary>
	public enum RECLS_RC
	{
			/// <summary> No search is currently active </summary> 
			SEARCH_NO_CURRENT          =   (-1 - 1001)
		,	/// <summary> The directory was invalid, or does not exist </summary> 
			INVALID_DIRECTORY          =   (-1 - 1002)
		,	/// <summary> No more data is available </summary> 
			NO_MORE_DATA               =   (-1 - 1003)
		,	/// <summary> Memory exhaustion </summary> 
			OUT_OF_MEMORY              =   (-1 - 1004)
		,	/// <summary> Function not implemented </summary> 
			NOT_IMPLEMENTED            =   (-1 - 1005)
		,	/// <summary> Invalid search type </summary> 
			INVALID_SEARCH_TYPE        =   (-1 - 1006)
		,	/// <summary> Invalid search protocol </summary> 
			INVALID_SEARCH_PROTOCOL    =   (-1 - 1007)
		,	/// <summary> Unexpected internal condition </summary> 
			UNEXPECTED                 =   (-1 - 1008)
		,	/// <summary> Cannot specify "." pattern in recursive search </summary> 
			DOT_RECURSIVE_SEARCH       =   (-1 - 1009)
		,	/// <summary> Search was cancelled by callback. It did not otherwise fail. </summary> 
			SEARCH_CANCELLED           =   (-1 - 1010)
		,	/// <summary> FTP infrastructure initialisation failed. </summary> 
			FTP_INIT_FAILED            =   (-1 - 1011)
		,	/// <summary> FTP connection failed. </summary> 
			FTP_CONNECTION_FAILED      =   (-1 - 1012)
		,	/// <summary> Given entry was of wrong type </summary> 
			WRONG_ENTRY_TYPE           =   (-1 - 1013)
		,	/// <summary> Cannot call stat() on root </summary> 
			CANNOT_STAT_ROOT           =   (-1 - 1014)
		,	/// <summary> Directory or pattern component exceeds maximum for host system </summary> 
			PATH_LIMIT_EXCEEDED         =   (-1 - 1015)
	};


	/// <summary>
	/// The exception type used by the recls API
	/// </summary>
	/// <remarks>Instances of this exception are throw when error conditions are encountered. They are 
	/// not thrown when a search returns no items, since that is a valid search result.</remarks>
	public class ReclsException
		: ApplicationException
	{
		/// <summary>
		/// Creates an instance of the exception
		/// </summary>
		internal ReclsException(string message, int code)
			: base(_make_x_str(message, code))
		{
			m_code = code;
		}

		private static string _make_x_str(string message, int code)
		{
			if(code == recls_api.RECLS_RC_OK)
			{
				return message;
			}
			else
			{
				return message + ": " + recls_api.GetErrorString(code) + " (" + code + ")";
			}
		}

		/// <summary>
		/// The recls error code associated with the exception
		/// </summary>
		public RECLS_RC Code
		{
			get { return (RECLS_RC)m_code; }
		}

		private readonly int m_code;
	}

	/// <summary>
	/// This class wraps access to the raw recls API
	/// </summary>
	internal sealed class recls_api
	{
		private recls_api()
		{}

		/// <summary>
		/// General success code
		/// </summary>
		public const int	RECLS_RC_OK						=	(0);
		/// <summary>
		/// General failure code
		/// </summary>
		public const int	RECLS_RC_FAIL                   =	(-1);

		/// <summary>
		/// No search is currently active
		/// </summary>
		public const int	RECLS_RC_SEARCH_NO_CURRENT      =	((-1 - 1001));
		/// <summary>
		/// The directory was invalid, or does not exist
		/// </summary>
		public const int	RECLS_RC_INVALID_DIRECTORY      =	((-1 - 1002));
		/// <summary>
		/// No more data is available
		/// </summary>
		public const int	RECLS_RC_NO_MORE_DATA           =	((-1 - 1003));
		/// <summary>
		/// Memory exhaustion
		/// </summary>
		public const int	RECLS_RC_OUT_OF_MEMORY          =	((-1 - 1004));
		/// <summary>
		/// Function not implemented
		/// </summary>
		public const int	RECLS_RC_NOT_IMPLEMENTED        =	((-1 - 1005));

		public static hrecls_t OpenSearch(string searchRoot, string pattern, RECLS_FLAG flags)
		{
			hrecls_t	hSrch;
			int 		code = Recls_Search(searchRoot, pattern, (uint)flags, out hSrch);

			if(code == RECLS_RC_NO_MORE_DATA)
			{
				hSrch = hrecls_t.Zero;
			}
			else if(code < 0)
			{
				throw new ReclsException("Failed to create search", code);
			}

			return hSrch;
		}

		public static hrecls_t OpenSearch(string searchRoot, string pattern, RECLS_FLAG flags, ProgressFnCallback progress)
		{
			hrecls_t	hSrch;
			int 		code = Recls_SearchFeedback(searchRoot, pattern, (uint)flags, progress, 0, out hSrch);

			if(code == RECLS_RC_NO_MORE_DATA)
			{
				hSrch = hrecls_t.Zero;
			}
			else if(code < 0)
			{
				throw new ReclsException("Failed to create search", code);
			}

			return hSrch;
		}

		public static hrecls_t OpenFtpSearch(string host, string username, string password, string searchRoot, string pattern, RECLS_FLAG flags)
		{
			hrecls_t	hSrch;
			int 		code = Recls_SearchFtp(host, username, password, searchRoot, pattern, (uint)flags, out hSrch);

			if(code == RECLS_RC_NO_MORE_DATA)
			{
				hSrch = hrecls_t.Zero;
			}
			else if(code < 0)
			{
				throw new ReclsException("Failed to create search", code);
			}

			return hSrch;
		}

		public static void CloseSearch(ref hrecls_t hSrch)
		{
			if(IntPtr.Zero != hSrch)
			{
				Recls_SearchClose(hSrch);
				hSrch = IntPtr.Zero;
			}
		}

		public static string GetErrorString(int code)
		{
			StringBuilder	buffer		=	new StringBuilder(261);
			uint			capacity	=	(uint)buffer.Capacity;
			int				cch 		= 	Recls_GetErrorString(code, buffer, capacity);

			buffer.Length = cch;

			return buffer.ToString();
		}

		public static void CloseEntry(ref recls_info_t entry)
		{
			if(IntPtr.Zero != entry)
			{
				Recls_CloseDetails(entry);
				entry = IntPtr.Zero;
			}
		}

		public static recls_info_t CopyEntry(recls_info_t entry)
		{
			recls_info_t	infoCopy;
			int				code	=	Recls_CopyDetails(entry, out infoCopy);

			if(code < 0)
			{
				throw new ReclsException("Failed to copy details", code);
			}

			return infoCopy;
		}

		public static bool NextEntry(hrecls_t hSrch, out recls_info_t entry)
		{
			int code = Recls_GetNextDetails(hSrch, out entry);

			if(code == RECLS_RC_NO_MORE_DATA)
			{
				return false;
			}
			else if(code < 0)
			{
				throw new ReclsException("Failed to get next entry", code);
			}
			else
			{
				return true;
			}
		}

		public static void GetEntry(hrecls_t hSrch, out recls_info_t entry)
		{
			int code = Recls_GetDetails(hSrch, out entry);

			if(code < 0)
			{
				throw new ReclsException("Failed to create search", code);
			}
		}

		public static recls_info_t Stat(string path, RECLS_FLAG flags)
		{
			recls_info_t	entry;
			int				code	=	Recls_Stat(path, (uint)flags, out entry);

			if(code < 0)
			{
				throw new ReclsException("Failed to stat() the requested path", code);
			}

			return entry;
		}

		/// <summary>
		/// A delegate defining the various string property access functions
		/// </summary>
		private delegate uint Recls_GetStringProperty_dg(recls_info_t fileInfo, StringBuilder buffer, uint cchBuffer);

		private static String GetEntryString(recls_info_t entry, Recls_GetStringProperty_dg dg)
		{
			uint			cch 		= 	dg(entry, null, 0);
			StringBuilder	buffer		=	new StringBuilder(1 + (int)cch);
			uint			capacity	=	(uint)buffer.Capacity;

			cch = dg(entry, buffer, capacity);

			buffer.Length = (int)cch;

			return buffer.ToString();
		}

		public static String GetEntryPath(recls_info_t entry)
		{
			return GetEntryString(entry, new Recls_GetStringProperty_dg(Recls_GetPathProperty));
		}

		public static char GetEntryDrive(recls_info_t entry)
		{
			char	chDrive;

			Recls_GetDriveProperty(entry, out chDrive);

			return chDrive;
		}

		public static String GetEntryDirectory(recls_info_t entry)
		{
			return GetEntryString(entry, new Recls_GetStringProperty_dg(Recls_GetDirectoryProperty));
		}

		public static String GetEntryDirectoryPath(recls_info_t entry)
		{
			return GetEntryString(entry, new Recls_GetStringProperty_dg(Recls_GetDirectoryPathProperty));
		}

		public static String GetEntryUNCDrive(recls_info_t entry)
		{
			return GetEntryString(entry, new Recls_GetStringProperty_dg(Recls_GetUNCDriveProperty));
		}

		public static String GetEntryFile(recls_info_t entry)
		{
			return GetEntryString(entry, new Recls_GetStringProperty_dg(Recls_GetFileProperty));
		}

		public static String GetEntryShortFile(recls_info_t entry)
		{
			return GetEntryString(entry, new Recls_GetStringProperty_dg(Recls_GetShortFileProperty));
		}

		public static String GetEntryFileName(recls_info_t entry)
		{
			return GetEntryString(entry, new Recls_GetStringProperty_dg(Recls_GetFileNameProperty));
		}

		public static String GetEntryFileExt(recls_info_t entry)
		{
			return GetEntryString(entry, new Recls_GetStringProperty_dg(Recls_GetFileExtProperty));
		}

		public static int GetEntryDirectoryPartsCount(recls_info_t entry)
		{
			return (int)Recls_GetDirectoryPartProperty(entry, -1, null, 0);
		}

		public static string GetEntryDirectoryPart(recls_info_t entry, int index)
		{
			StringBuilder	buffer		=	new StringBuilder(261);
			uint			capacity	=	(uint)buffer.Capacity;
			uint			cch 		= 	Recls_GetDirectoryPartProperty(entry, index, buffer, capacity);

			buffer.Length = (int)cch;

			return buffer.ToString();
		}

		public static DateTime GetEntryCreationTime(recls_info_t entry)
		{
			return DateTime.FromFileTime(Recls_GetCreationTime(entry));
		}
		public static DateTime GetEntryModificationTime(recls_info_t entry)
		{
			return DateTime.FromFileTime(Recls_GetModificationTime(entry));
		}
		public static DateTime GetEntryLastAccessTime(recls_info_t entry)
		{
			return DateTime.FromFileTime(Recls_GetLastAccessTime(entry));
		}
		public static DateTime GetEntryLastStatusChangeTime(recls_info_t entry)
		{
			return DateTime.FromFileTime(Recls_GetLastStatusChangeTime(entry));
		}

		public static ulong GetEntrySize(recls_info_t entry)
		{
			recls_filesize_t    size;

			Recls_GetSizeProperty(entry, out size);

			return size;
		}

		public static bool GetEntryIsReadOnly(recls_info_t entry)
		{
			return 0 != Recls_IsFileReadOnly(entry);
		}

		public static bool GetEntryIsDirectory(recls_info_t entry)
		{
			return 0 != Recls_IsFileDirectory(entry);
		}

		public static bool GetEntryIsLink(recls_info_t entry)
		{
			return 0 != Recls_IsFileLink(entry);
		}

		public static bool GetEntryIsUNC(recls_info_t entry)
		{
			return 0 != Recls_IsFileUNC(entry);
		}

		[DllImport(DLL_NAME, EntryPoint = "Recls_SearchFeedback", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_SearchFeedback_(string searchRoot, string pattern, uint flags, ProgressFnCallback pfn, int param, out hrecls_t hSrch);

		private static int Recls_SearchFeedback(string searchRoot, string pattern, uint flags, ProgressFnCallback pfn, int param, out hrecls_t hSrch)
		{
			uint	RECLS_F_CALLBACKS_STDCALL_ON_WIN32	=	0x01000000;

			return Recls_SearchFeedback_(searchRoot, pattern, flags | RECLS_F_CALLBACKS_STDCALL_ON_WIN32, pfn, param, out hSrch);
		}

		/// <summary>
		/// Starts a search in the given root, for the given pattern according to the given flags
		/// </summary>
		[DllImport(DLL_NAME, EntryPoint = "Recls_Search", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_Search(string searchRoot, string pattern, uint flags, out hrecls_t hSrch);

		[DllImport(DLL_NAME, EntryPoint = "Recls_SearchFtp", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_SearchFtp(string host, string username, string password, string searchRoot, string pattern, uint flags, out hrecls_t hSrch);

		[DllImport(DLL_NAME, EntryPoint = "Recls_SearchClose", CallingConvention=CallingConvention.Cdecl, ExactSpelling=true)]
		private static extern void Recls_SearchClose(hrecls_t hSrch);

		[DllImport(DLL_NAME, EntryPoint = "Recls_GetErrorString", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_GetErrorString(int code, StringBuilder buffer, uint cchBuffer);

		[DllImport(DLL_NAME, EntryPoint = "Recls_GetNextDetails", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_GetNextDetails(hrecls_t hSrch, out recls_info_t info);

		[DllImport(DLL_NAME, EntryPoint = "Recls_GetDetails", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_GetDetails(hrecls_t hSrch, out recls_info_t info);

		[DllImport(DLL_NAME, EntryPoint = "Recls_CopyDetails", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_CopyDetails(recls_info_t fileInfo, out recls_info_t infoCopy);

		[DllImport(DLL_NAME, EntryPoint = "Recls_CloseDetails", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_CloseDetails(recls_info_t fileInfo);

		[DllImport(DLL_NAME, EntryPoint = "Recls_GetPathProperty", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern uint Recls_GetPathProperty(recls_info_t fileInfo, StringBuilder buffer, uint cchBuffer);

		[DllImport(DLL_NAME, EntryPoint = "Recls_GetDriveProperty", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern void Recls_GetDriveProperty(recls_info_t fileInfo, out char chDrive);

		[DllImport(DLL_NAME, EntryPoint = "Recls_GetDirectoryProperty", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern uint Recls_GetDirectoryProperty(recls_info_t fileInfo, StringBuilder buffer, uint cchBuffer);

		[DllImport(DLL_NAME, EntryPoint = "Recls_GetDirectoryPathProperty", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern uint Recls_GetDirectoryPathProperty(recls_info_t fileInfo, StringBuilder buffer, uint cchBuffer);

		[DllImport(DLL_NAME, EntryPoint = "Recls_GetUNCDriveProperty", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern uint Recls_GetUNCDriveProperty(recls_info_t fileInfo, StringBuilder buffer, uint cchBuffer);

		[DllImport(DLL_NAME, EntryPoint = "Recls_GetShortFileProperty", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern uint Recls_GetShortFileProperty(recls_info_t fileInfo, StringBuilder buffer, uint cchBuffer);

		[DllImport(DLL_NAME, EntryPoint = "Recls_GetFileProperty", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern uint Recls_GetFileProperty(recls_info_t fileInfo, StringBuilder buffer, uint cchBuffer);

		[DllImport(DLL_NAME, EntryPoint = "Recls_GetFileNameProperty", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern uint Recls_GetFileNameProperty(recls_info_t fileInfo, StringBuilder buffer, uint cchBuffer);

		[DllImport(DLL_NAME, EntryPoint = "Recls_GetFileExtProperty", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern uint Recls_GetFileExtProperty(recls_info_t fileInfo, StringBuilder buffer, uint cchBuffer);

		[DllImport(DLL_NAME, EntryPoint = "Recls_GetCreationTime", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern long Recls_GetCreationTime(recls_info_t fileInfo);
		[DllImport(DLL_NAME, EntryPoint = "Recls_GetModificationTime", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern long Recls_GetModificationTime(recls_info_t fileInfo);
		[DllImport(DLL_NAME, EntryPoint = "Recls_GetLastAccessTime", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern long Recls_GetLastAccessTime(recls_info_t fileInfo);
		[DllImport(DLL_NAME, EntryPoint = "Recls_GetLastStatusChangeTime", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern long Recls_GetLastStatusChangeTime(recls_info_t fileInfo);

		[DllImport(DLL_NAME, EntryPoint = "Recls_GetSizeProperty", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern void Recls_GetSizeProperty(recls_info_t fileInfo, out recls_filesize_t size);

		[DllImport(DLL_NAME, EntryPoint = "Recls_GetDirectoryPartProperty", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern uint Recls_GetDirectoryPartProperty(recls_info_t fileInfo, int part, StringBuilder buffer, uint cchBuffer);

		[DllImport(DLL_NAME, EntryPoint = "Recls_IsFileReadOnly", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_IsFileReadOnly(recls_info_t fileInfo);

		[DllImport(DLL_NAME, EntryPoint = "Recls_IsFileDirectory", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_IsFileDirectory(recls_info_t fileInfo);

		[DllImport(DLL_NAME, EntryPoint = "Recls_IsFileLink", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_IsFileLink(recls_info_t fileInfo);

		[DllImport(DLL_NAME, EntryPoint = "Recls_IsFileUNC", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_IsFileUNC(recls_info_t fileInfo);

		[DllImport(DLL_NAME, EntryPoint = "Recls_Stat", CallingConvention=CallingConvention.Cdecl, CharSet=CharSet.Ansi, ExactSpelling=true)]
		private static extern int Recls_Stat(string searchRoot, uint flags, out recls_info_t entry);

#if DEBUG
		private const string DLL_NAME	=	"recls.1.debug.dll";
#else // ? DEBUG
		private const string DLL_NAME	=	"recls.1.dll";
#endif // DEBUG
	}

	// This class serves as the namespace summary doc for NDoc, because C# cannot use <summary> tags on namespaces
	/// <summary>
	/// Contains the recls C# mapping classes and enumerations
	/// </summary>
	class NamespaceDoc
	{}
}
