/* /////////////////////////////////////////////////////////////////////////////
 * File:        ReclsFileSearchDirectoryNode.cpp
 *
 * Purpose:     Implementation of the ReclsFileSearchDirectoryNode class.
 *
 * Created:     31st May 2004
 * Updated:     27th December 2005
 *
 * Home:        http://recls.org/
 *
 * Copyright (c) 2004-2005, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#include <recls/recls.h>
#include <recls/assert.h>
#include "recls_string.h"
#include "recls_util.h"
#include "recls_impl.h"

#if defined(RECLS_CHAR_TYPE_IS_WCHAR)
# if defined(RECLS_PLATFORM_IS_WIN32)
#  include <winstl/char_conversions.hpp>
# endif /* RECLS_PLATFORM_IS_WIN32 */
#endif /* RECLS_CHAR_TYPE_IS_???? */

#include "recls_debug.h"

#include "ReclsFileSearchDirectoryNode.hpp"
#include "EntryFunctions.h"
#include "entryinfo.hpp"

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{
#endif /* !RECLS_NO_NAMESPACE */

namespace {

typedef ReclsFileSearchDirectoryNode::traits_type               traits_type;
typedef ReclsFileSearchDirectoryNode::file_path_buffer_type     file_path_buffer_type;

stlsoft_ns_using(c_str_ptr)
stlsoft_ns_using(c_str_len)

} // anonymous namespace

/* /////////////////////////////////////////////////////////////////////////////
 * ReclsFileSearchDirectoryNode
 */

/* static */ int ReclsFileSearchDirectoryNode::essFlags_from_reclsFlags_(recls_uint32_t flags)
{
    function_scope_trace("ReclsFileSearchDirectoryNode::essFlags_from_reclsFlags_");

#ifdef _RECLS_USING_STLSOFT_SEARCHSPEC_SEQUENCE
    // Because Digital Mars 8.40- has a problem, we must access the typedef separately from the enum value
    typedef entry_sequence_type::find_sequence_type sequence_t;
#else /* ? _RECLS_USING_STLSOFT_SEARCHSPEC_SEQUENCE */
    typedef entry_sequence_type                     sequence_t;
#endif /* _RECLS_USING_STLSOFT_SEARCHSPEC_SEQUENCE */

    int ssFlags = 0;

    if(0 != (flags & RECLS_F_FILES))
    {
        ssFlags |= sequence_t::files;
    }
    if(0 != (flags & RECLS_F_DIRECTORIES))
    {
        ssFlags |= sequence_t::directories;
    }

    return ssFlags;
}

/* static */ int ReclsFileSearchDirectoryNode::dssFlags_from_reclsFlags_(recls_uint32_t flags)
{
    function_scope_trace("ReclsFileSearchDirectoryNode::dssFlags_from_reclsFlags_");

    typedef directory_sequence_type                 sequence_t;

    int ssFlags = 0;

    ssFlags |= sequence_t::directories;

#if defined(RECLS_PLATFORM_IS_UNIX)
# ifdef __SYNSOFT_DBS_COMPILER_SUPPORTS_PRAGMA_MESSAGE
#  pragma message(_sscomp_fileline_message("TODO: Make this for all, once findfile_sequence supports fullPath"))
# endif /* __SYNSOFT_DBS_COMPILER_SUPPORTS_PRAGMA_MESSAGE */

    ssFlags |= sequence_t::fullPath;
#endif /* platform */

    if(0 == (flags & RECLS_F_ALLOW_REPARSE_DIRS))
    {
#if defined(RECLS_PLATFORM_IS_WIN32)
        ssFlags |= sequence_t::skipReparseDirs;
#endif /* platform */
    }

    return ssFlags;
}

/* static */ ReclsFileSearchDirectoryNode::directory_sequence_type::const_iterator ReclsFileSearchDirectoryNode::select_iter_if_(unsigned long b, directory_sequence_type::const_iterator trueVal, directory_sequence_type::const_iterator falseVal)
{
    function_scope_trace("ReclsFileSearchDirectoryNode::select_iter_if_");

    // I can't explain it, but Borland does not like the tertiary operator and the copy-ctors of the iterators
    if(b)
    {
        return trueVal;
    }
    else
    {
        return falseVal;
    }
}

/* static */ size_t ReclsFileSearchDirectoryNode::prepare_searchDir_(file_path_buffer_type &buff, recls_char_t const *searchDir)
{
    function_scope_trace("ReclsFileSearchDirectoryNode::prepare_searchDir_");

    size_t  len =   traits_type::str_len(searchDir);

    buff[len] = '\0';
    traits_type::str_n_copy(&buff[0], searchDir, len);

    RECLS_MESSAGE_ASSERT("I _think_ this should never happen. Please report if it does.", 0 != len);

    if( 0 != len &&
        !traits_type::has_dir_end(&buff[0] + (len - 1)))
    {
        traits_type::ensure_dir_end(&buff[0] + (len - 1));
        ++len;
    }

    return len;
}

/* static */ recls_info_t ReclsFileSearchDirectoryNode::CreateEntryInfo(size_t                                                              rootDirLen
                                                                    ,   recls_char_t const                                                  *searchDir
                                                                    ,   size_t                                                              searchDirLen
                                                                    ,   recls_uint32_t                                                      flags
                                                                    ,   ReclsFileSearchDirectoryNode::entry_sequence_type::const_iterator   it)
{
    function_scope_trace("ReclsFileSearchDirectoryNode::CreateEntryInfo");

#if defined(RECLS_PLATFORM_IS_UNIX)

    typedef int (*PfnStat)(char const *, struct stat *);

# if defined(WIN32) || \
  defined(_WIN32)
    PfnStat             pfn =   ::stat;
# else /* ? WIN32 */
    PfnStat             pfn =   (RECLS_F_LINKS == (flags & RECLS_F_LINKS)) ? ::lstat : ::stat;
# endif /* WIN32 */
    struct stat         st;
    recls_char_t const  *entryPath  =   *it;

    if(0 != (*pfn)(entryPath, &st))
    {
        // This will cause RECLS_F_OUT_OF_MEMORY. 
        // TODO: Fix it!
        return NULL;
    }
    else
    {
        size_t const        entryPathLen    =   traits_type::str_len(entryPath);
        recls_char_t const  *entryFile      =   traits_type::str_rchr(&entryPath[0], traits_type::path_name_separator()) + 1;
        RECLS_ASSERT(NULL != (entryFile - 1));
        size_t const        entryFileLen    =   entryPathLen - (entryFile - entryPath);
        RECLS_ASSERT(entryFileLen == traits_type::str_len(entryFile));

        return create_entryinfo(rootDirLen, searchDir, searchDirLen, entryPath, entryPathLen, entryFile, entryFileLen, flags, &st);
    }
#elif defined(RECLS_PLATFORM_IS_WIN32)
    // In this case:
    //
    // - searchDir contains the directory that was searched
    // - it contains the iterator from which the full path and the stat data may be obtained.
    //    Note: the full path may contain additional subdirectories over and above that of searchDir
#if 0
    ReclsFileSearchDirectoryNode::entry_sequence_type::value_type const &value  =   *it;
#else /* ? 0 */
    ReclsFileSearchDirectoryNode::entry_sequence_type::const_reference  value   =   *it;
#endif /* 0 */

    recls_char_t const  *entryPath      =   value.get_path();
    size_t const        entryPathLen    =   traits_type::str_len(entryPath);
    recls_char_t const  *entryFile      =   traits_type::str_rchr(&entryPath[0], traits_type::path_name_separator()) + 1;
    RECLS_ASSERT(NULL != (entryFile - 1));
    size_t const        entryFileLen    =   entryPathLen - (entryFile - entryPath);
    RECLS_ASSERT(entryFileLen == traits_type::str_len(entryFile));

    return create_entryinfo(rootDirLen, searchDir, searchDirLen, entryPath, entryPathLen, entryFile, entryFileLen, flags, &value.get_find_data());
#else /* ? platform */
# error Platform not discriminated
#endif /* platform */
}

ReclsFileSearchDirectoryNode::ReclsFileSearchDirectoryNode( recls_uint32_t              flags
                                                        ,   size_t                      rootDirLen
                                                        ,   recls_char_t const          *searchDir
                                                        ,   recls_char_t const          *pattern
                                                        ,   hrecls_progress_fn_t        pfn
                                                        ,   recls_process_fn_param_t    param)
    : m_current(NULL)
    , m_dnode(NULL)
    , m_flags(flags)
    , m_rootDirLen(rootDirLen)
    , m_searchDir()
    , m_searchDirLen(prepare_searchDir_(m_searchDir, searchDir))
    , m_pattern(pattern)
    , m_directories(searchDir
#if defined(RECLS_PLATFORM_IS_WIN32)    // Windows uses findfile_sequence, which takes wildcards
                ,   traits_type::pattern_all()
#endif /* platform */
                ,   dssFlags_from_reclsFlags_(flags))
    , m_directoriesBegin(select_iter_if_((flags & RECLS_F_RECURSIVE), m_directories.begin(), m_directories.end()))
    , m_entries(searchDir
            ,   pattern
#ifdef _RECLS_SUPPORTS_MULTIPATTERN
            ,   traits_type::path_separator()
#endif /* _RECLS_SUPPORTS_MULTIPATTERN */
            ,   essFlags_from_reclsFlags_(flags))
    , m_entriesBegin(m_entries.begin())
    , m_pfn(pfn)
    , m_param(param)
{
    function_scope_trace("ReclsFileSearchDirectoryNode::ReclsFileSearchDirectoryNode");

    RECLS_ASSERT(0 != m_searchDirLen);
    RECLS_ASSERT(traits_type::is_path_name_separator(m_searchDir[m_searchDirLen - 1]));

    RECLS_ASSERT(STLSOFT_RAW_OFFSETOF(ReclsFileSearchDirectoryNode, m_entries) < STLSOFT_RAW_OFFSETOF(ReclsFileSearchDirectoryNode, m_entriesBegin));
    RECLS_ASSERT(STLSOFT_RAW_OFFSETOF(ReclsFileSearchDirectoryNode, m_directories) < STLSOFT_RAW_OFFSETOF(ReclsFileSearchDirectoryNode, m_directoriesBegin));

#if !defined(STLSOFT_COMPILER_IS_COMO) && \
    !defined(STLSOFT_COMPILER_IS_GCC) && \
    !defined(STLSOFT_COMPILER_IS_INTEL)
    STLSOFT_STATIC_ASSERT(STLSOFT_RAW_OFFSETOF(ReclsFileSearchDirectoryNode, m_entries) < STLSOFT_RAW_OFFSETOF(ReclsFileSearchDirectoryNode, m_entriesBegin));
    STLSOFT_STATIC_ASSERT(STLSOFT_RAW_OFFSETOF(ReclsFileSearchDirectoryNode, m_directories) < STLSOFT_RAW_OFFSETOF(ReclsFileSearchDirectoryNode, m_directoriesBegin));
#endif /* compiler */
}

/* static */ ReclsFileSearchDirectoryNode *ReclsFileSearchDirectoryNode::FindAndCreate( recls_uint32_t              flags
                                                                                    ,   size_t                      rootDirLen
                                                                                    ,   recls_char_t const          *searchDir
                                                                                    ,   recls_char_t const          *pattern
                                                                                    ,   hrecls_progress_fn_t        pfn
                                                                                    ,   recls_process_fn_param_t    param)
{
    ReclsFileSearchDirectoryNode  *node;

    function_scope_trace("ReclsFileSearchDirectoryNode::FindAndCreate");

#ifdef RECLS_COMPILER_THROWS_ON_NEW_FAIL
    try
    {
#endif /* RECLS_COMPILER_THROWS_ON_NEW_FAIL */
        node = new ReclsFileSearchDirectoryNode(flags, rootDirLen, searchDir, pattern, pfn, param);
#ifdef RECLS_COMPILER_THROWS_ON_NEW_FAIL
    }
    catch(std::bad_alloc &)
    {
        node = NULL;
    }
#endif /* RECLS_COMPILER_THROWS_ON_NEW_FAIL */

    if(NULL != node)
    {
        // Ensure that it, or one of its sub-nodes, has matching entries.
        recls_rc_t  rc = node->Initialise();

        if(RECLS_FAILED(rc))
        {
            delete node;

            node = NULL;
        }
    }

    RECLS_ASSERT(NULL == node || node->is_valid());

    return node;
}

/* static */ recls_rc_t ReclsFileSearchDirectoryNode::Stat( recls_char_t const  *path
                                                        ,   recls_uint32_t      flags
                                                        ,   recls_info_t        *phEntry)
{
    function_scope_trace("ReclsFileSearchDirectoryNode::Stat");

    // 1. Ensure that the path is valid
    // 
    // 1.a Must not be NULL
    if(NULL == path)
    {
        return RECLS_RC_INVALID_DIRECTORY;
    }

#if 0
fprintf(stdout, "[path=%s]\n", path);
#endif /* 0 */

    // 
    // 1.b Must not be > max_path()
    const size_t    pathLen =   traits_type::str_len(path);

    if(pathLen > traits_type::path_max())
    {
        return RECLS_RC_PATH_LIMIT_EXCEEDED;
    }

    // 1. c Ensure that is has the correct slashes
    file_path_buffer_type   path_;

    path = traits_type::str_copy(&path_[0], path);

#if defined(RECLS_PLATFORM_IS_UNIX) && \
    defined(WIN32) && \
    defined(EMULATE_UNIX_ON_WIN32)
    // UNIX
    stlsoft_ns_qual_std(replace)(&path_[0], &path_[0] + path_.size(), RECLS_LITERAL('\\'), RECLS_LITERAL('/'));
#elif defined(RECLS_PLATFORM_IS_WIN32)
    stlsoft_ns_qual_std(replace)(&path_[0], &path_[0] + path_.size(), RECLS_LITERAL('/'), RECLS_LITERAL('\\'));
# endif /* WIN32 && EMULATE_UNIX_ON_WIN32 */

    // 2. Check that we're not stat()-ing the root directory, since that's 
    //     not supported
    if(0 == traits_type::str_compare(RECLS_LITERAL("/"), path))
    {
        return RECLS_RC_CANNOT_STAT_ROOT;
    }
#if defined(RECLS_PLATFORM_IS_WIN32) || \
    (   defined(RECLS_PLATFORM_IS_UNIX) && \
        defined(WIN32) && \
        defined(EMULATE_UNIX_ON_WIN32))
    if(0 == traits_type::str_compare(RECLS_LITERAL("\\"), path))
    {
        return RECLS_RC_CANNOT_STAT_ROOT;
    }
    if( isalpha(path[0]) &&
        (   0 == traits_type::str_compare(RECLS_LITERAL(":\\"), path + 1) ||
            0 == traits_type::str_compare(RECLS_LITERAL(":/"), path + 1)))
    {
        return RECLS_RC_CANNOT_STAT_ROOT;
    }
#endif /* Windows root paths */

    // Now need to remove the directory end, if any
    traits_type::remove_dir_end(&path_[0]);

    traits_type::stat_data_type     st;
    traits_type::stat_data_type     *pst    =   &st;

    if(!traits_type::stat(path, &st))
    {
        pst = NULL;
    }

    if( NULL == pst &&
        RECLS_F_DETAILS_LATER != (flags & (RECLS_F_DETAILS_LATER | RECLS_F_TYPEMASK))) // To allow non-existant things to be stat'd
    {
        return RECLS_RC_NO_MORE_DATA;
    }
    else
    {
        size_t const        pathLen         =   traits_type::str_len(path);
        recls_char_t const  *entryFile      =   traits_type::str_rchr(path, traits_type::path_name_separator()) + 1;
        RECLS_ASSERT(NULL != (entryFile - 1));
        size_t const        entryDirLen     =   static_cast<size_t>(entryFile - path);
        size_t const        entryFileLen    =   pathLen - entryDirLen;
        RECLS_ASSERT(entryFileLen == traits_type::str_len(entryFile));

#if 0
fprintf(stdout
    ,   "[create_entryinfo(%.*s [%d], %.*s [%d], %.*s [%d], 0x%08x]\n"
    ,   entryDirLen, path, entryDirLen
    ,   pathLen, path, pathLen
    ,   entryFileLen, entryFile, entryFileLen
    ,   flags
        );
#endif /* 0 */

        *phEntry = create_entryinfo(pathLen, path, entryDirLen, path, pathLen, entryFile, entryFileLen, flags, pst);

        return RECLS_RC_OK;
    }

#if defined(STLSOFT_COMPILER_IS_GCC)
    RECLS_MESSAGE_ASSERT("Should never reach here", 0);

    return RECLS_RC_OK;
#endif /* GCC */
}

ReclsFileSearchDirectoryNode::~ReclsFileSearchDirectoryNode()
{
    function_scope_trace("ReclsFileSearchDirectoryNode::~ReclsFileSearchDirectoryNode");

    Entry_Release(m_current);

    delete m_dnode;
}

recls_rc_t ReclsFileSearchDirectoryNode::Initialise()
{
    function_scope_trace("ReclsFileSearchDirectoryNode::Initialise");

    recls_rc_t  rc;

    RECLS_ASSERT(NULL == m_current);
    RECLS_ASSERT(NULL == m_dnode);

    if(NULL != m_pfn)
    {
#if defined(RECLS_PLATFORM_IS_WIN32)
        if(m_flags & RECLS_F_CALLBACKS_STDCALL_ON_WIN32)
        {
            typedef int (RECLS_CALLCONV_STDDECL *stdcall_progress_fn_t)(recls_char_t const          *dir
                                                                    ,   size_t                      dirLen
                                                                    ,   recls_process_fn_param_t    param
                                                                    ,   void                        *reserved0
                                                                    ,   recls_uint32_t              reserved1);

            stdcall_progress_fn_t   pfn_stdcall =   (stdcall_progress_fn_t)m_pfn;

            (*pfn_stdcall)(m_searchDir.c_str(), m_searchDirLen, m_param, NULL, 0);
        }
        else
#endif /* RECLS_PLATFORM_IS_WIN32 */
        {
            if(0 == (*m_pfn)(m_searchDir.c_str(), m_searchDirLen, m_param, NULL, 0))
            {
                return RECLS_RC_USER_CANCELLED_SEARCH;
            }
        }
    }

    if(m_entriesBegin != m_entries.end())
    {
# if defined(RECLS_CHAR_TYPE_IS_CHAR)
        debug_printf("Next entry in %s\n", static_cast<char const*>(c_str_ptr(m_searchDir)));
# elif defined(RECLS_CHAR_TYPE_IS_WCHAR)
        debug_printf("Next entry in %s\n", static_cast<char const*>(winstl::w2a(m_searchDir)));
# endif /* RECLS_CHAR_TYPE_IS_???? */

        // (i) Try getting a file first,
        m_current = CreateEntryInfo(m_rootDirLen, c_str_ptr(m_searchDir), m_searchDirLen, m_flags, m_entriesBegin);

        if(NULL == m_current)
        {
            rc = RECLS_RC_OUT_OF_MEMORY;
        }
        else
        {
            rc = RECLS_RC_OK;
        }
    }
    else
    {
        if(m_directoriesBegin == m_directories.end())
        {
            rc = RECLS_RC_NO_MORE_DATA;
        }
        else
        {
//          debug_printf("Next directory in %s\n", static_cast<char const*>(c_str_ptr(m_searchDir)));
//          debug_printf("*m_directoriesBegin: %s\n", static_cast<char const*>(c_str_ptr(*m_directoriesBegin)));

            do
            {
#ifdef __SYNSOFT_DBS_COMPILER_SUPPORTS_PRAGMA_MESSAGE
# pragma message(_sscomp_fileline_message("TODO: Merge these once we've got findfile_sequence supporting fullPath"))
#endif /* __SYNSOFT_DBS_COMPILER_SUPPORTS_PRAGMA_MESSAGE */

//              RECLS_ASSERT('\0' != (*m_directoriesBegin).get_path()[0]);

                m_dnode = ReclsFileSearchDirectoryNode::FindAndCreate(  m_flags
                                                                    ,   m_rootDirLen
#if defined(RECLS_PLATFORM_IS_UNIX)
                // The way glob_sequence works
                                                                    ,   *m_directoriesBegin
#elif defined(RECLS_PLATFORM_IS_WIN32)
                // The way basic_findfile_sequence<> works
                                                                    ,   (*m_directoriesBegin).get_path()
#endif /* RECLS_PLATFORM_IS_??? */
                                                                    ,   c_str_ptr(m_pattern)
                                                                    ,   m_pfn
                                                                    ,   m_param);

            } while(NULL == m_dnode && ++m_directoriesBegin != m_directories.end());

            rc = (NULL == m_dnode) ? RECLS_RC_NO_MORE_DATA : RECLS_RC_OK;
        }
    }

    if(RECLS_SUCCEEDED(rc))
    {
        RECLS_ASSERT(is_valid());
    }

    return rc;
}

#ifdef RECLS_ENFORCING_CONTRACTS
recls_bool_t ReclsFileSearchDirectoryNode::is_valid() const
{
    function_scope_trace("ReclsFileSearchDirectoryNode::is_valid");

    recls_rc_t  rc  =   RECLS_RC_OK;

#if defined(RECLS_PRAGMA_MESSAGE_SUPPORT) && \
    !defined(RECLS_PRAGMA_MESSAGE_IS_INTRUSIVE)
# pragma message("Flesh these out")
    if(RECLS_SUCCEEDED(rc))
    {

    }
#endif /* compiler */

    // (i) Either we are enumerating files (m_current != NULL) or directories (m_dnode != NULL), but not both
    RECLS_ASSERT(NULL == m_current || NULL == m_dnode);
    // (ii) Either we are enumerating files (m_current != NULL) or there are no more files to enumerate
    RECLS_ASSERT(NULL != m_current || m_entriesBegin == m_entries.end());

    return RECLS_SUCCEEDED(rc);
}
#endif /* RECLS_ENFORCING_CONTRACTS */

recls_rc_t ReclsFileSearchDirectoryNode::GetNext()
{
    function_scope_trace("ReclsFileSearchDirectoryNode::GetNext");

    RECLS_ASSERT(is_valid());

    /* Searching operates as follows:
     *
     * 1. Return all the contents of the files
     * 2. Return the contents of the directories.
     *
     * Hence, if m_dnode is non-NULL, we've already searched
     */

    /* States:
     *
     * - Iterating files: m_entriesBegin != m_entries.end(), m_dnode is NULL, m_directoriesBegin != m_directories.end()
     * - Iterating directories: m_directoriesBegin != m_directories.end(), m_dnode is non-NULL, m_current is NULL
     *
     */

    // Invariants

    // (i) Either we are enumerating files (m_current != NULL) or directories (m_dnode != NULL), but not both
    RECLS_ASSERT(NULL == m_current || NULL == m_dnode);
    // (ii) Either we are enumerating files (m_current != NULL) or there are no more files to enumerate
    RECLS_ASSERT(NULL != m_current || m_entriesBegin == m_entries.end());

    recls_rc_t  rc = RECLS_RC_NO_MORE_DATA;

    if(NULL != m_current)
    {
        // Currently enumerating through the files
        RECLS_ASSERT(m_entriesBegin != m_entries.end());
        RECLS_ASSERT(NULL == m_dnode);

        // Advance, and check for end of sequence
        ++m_entriesBegin;

        Entry_Release(m_current);
        if(m_entriesBegin != m_entries.end())
        {
            // Still enumerating, so just update m_current
            m_current = CreateEntryInfo(m_rootDirLen, c_str_ptr(m_searchDir), m_searchDirLen, m_flags, m_entriesBegin);

            rc = RECLS_RC_OK;
        }
        else
        {
            // No more left in the files sequence, so delete m_current
            m_current = NULL;

            rc = RECLS_RC_NO_MORE_DATA;
        }
    }

    if(NULL == m_current)
    {
        // Now we are either enumerating the directories, or we've already done so
        if(NULL != m_dnode)
        {
            // Currently enumerating the directories
            rc = m_dnode->GetNext();

            if(RECLS_RC_NO_MORE_DATA == rc)
            {
                ++m_directoriesBegin;

                delete m_dnode;

                m_dnode = NULL;
            }
        }

        if(m_directoriesBegin == m_directories.end())
        {
            // Enumeration is complete.
            rc = RECLS_RC_NO_MORE_DATA;
        }
        else
        {
            if(NULL == m_dnode)
            {
                do
                {
                    // Creation of the node will cause it to enter the first enumeration
                    // state. However, if there are no matching, then NULL will be returned
                    RECLS_ASSERT(m_directoriesBegin != m_directories.end());

                    m_dnode = ReclsFileSearchDirectoryNode::FindAndCreate(  m_flags
                                                                        ,   m_rootDirLen
#if defined(RECLS_PLATFORM_IS_UNIX)
                                                                        ,   *m_directoriesBegin
#elif defined(RECLS_PLATFORM_IS_WIN32)
                                                                        ,   (*m_directoriesBegin).get_path()
#endif /* RECLS_PLATFORM_IS_??? */
                                                                        ,   c_str_ptr(m_pattern)
                                                                        ,   m_pfn
                                                                        ,   m_param);

                    if(NULL != m_dnode)
                    {
                        rc = RECLS_RC_OK;
                    }
                    else
                    {
                        ++m_directoriesBegin;
                    }

                } while(NULL == m_dnode && m_directoriesBegin != m_directories.end());
            }
        }
    }

    RECLS_ASSERT(is_valid());

    return rc;
}

recls_rc_t ReclsFileSearchDirectoryNode::GetDetails(recls_info_t *pinfo)
{
    function_scope_trace("ReclsFileSearchDirectoryNode::GetDetails");

    RECLS_ASSERT(is_valid());

    recls_rc_t  rc;

    RECLS_ASSERT(NULL != pinfo);
    RECLS_ASSERT(NULL == m_current || NULL == m_dnode);

    if(NULL != m_current)
    {
        // Currently searching for files from the current directory

        RECLS_ASSERT(NULL == m_dnode);

        rc = Entry_Copy(m_current, pinfo);

#if defined(_RECLS_DEBUG)
        {
            file_path_buffer_type   buffer;

            Recls_GetPathProperty(m_current, &buffer[0], buffer.size());

# if defined(RECLS_CHAR_TYPE_IS_CHAR)
            debug_printf("    [%s]\n", static_cast<char const*>(c_str_ptr(buffer)));
# elif defined(RECLS_CHAR_TYPE_IS_WCHAR)
            debug_printf("    [%s]\n", static_cast<char const*>(winstl::w2a(buffer)));
# endif /* RECLS_CHAR_TYPE_IS_???? */
        }
#endif /* _RECLS_DEBUG */
    }
    else if(NULL != m_dnode)
    {
        RECLS_ASSERT(NULL == m_current);

        // Sub-directory searching is active, so get from there.
        rc = m_dnode->GetDetails(pinfo);
    }
    else
    {
        // Enumeration has completed
        rc = RECLS_RC_NO_MORE_DATA;
    }

    RECLS_ASSERT(is_valid());

    return rc;
}

recls_rc_t ReclsFileSearchDirectoryNode::GetNextDetails(recls_info_t *pinfo)
{
    function_scope_trace("ReclsFileSearchDirectoryNode::GetNextDetails");

    RECLS_ASSERT(is_valid());
    RECLS_ASSERT(NULL != pinfo);

    recls_rc_t  rc  =   GetNext();

    if(RECLS_SUCCEEDED(rc))
    {
        rc = GetDetails(pinfo);
    }

    RECLS_ASSERT(is_valid());

    return rc;
}

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */
