/* /////////////////////////////////////////////////////////////////////////////
 * File:        ReclsFileSearchDirectoryNode.h
 *
 * Purpose:     ReclsFileSearchDirectoryNode class.
 *
 * Created:     31st May 2004
 * Updated:     26th December 2005
 *
 * Home:        http://recls.org/
 *
 * Copyright (c) 2004-2005, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name(s) of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/** \file ReclsFileSearchDirectoryNode.h Definition of the ReclsFileSearchDirectoryNode class
 */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

// recls includes
#include <recls/recls.h>
#include "recls_string.h"
#include "recls_util.h"

// Platform-specific includes
#include <platformstl/platformstl.h>
#include <platformstl/filesystem_traits.hpp>
#include <platformstl/file_path_buffer.hpp>

#if defined(RECLS_PLATFORM_IS_UNIX)

# include <unixstl/glob_sequence.hpp>
# include <unixstl/readdir_sequence.hpp>

#elif defined(RECLS_PLATFORM_IS_WIN32)

# include <winstl/findfile_sequence.hpp>

#else /* unrecognised platform */

# error The platform is not recognised. Only RECLS_PLATFORM_IS_UNIX or RECLS_PLATFORM_IS_WIN32 currently supported

#endif /* platform */

// STLSoft includes
#ifdef _RECLS_SUPPORTS_MULTIPATTERN
# if !defined(RECLS_PLATFORM_IS_WIN32) || \
     !defined(WINSTL_VER_WINSTL_HPP_FINDFILE_SEQUENCE_MAJOR) || \
     WINSTL_VER_WINSTL_HPP_FINDFILE_SEQUENCE_MAJOR < 3
#  define _RECLS_USING_STLSOFT_SEARCHSPEC_SEQUENCE
#  include <stlsoft/searchspec_sequence.hpp>
# endif /* stlsoft::findfile_sequence version */
#endif /* _RECLS_SUPPORTS_MULTIPATTERN */

#include "ReclsSearch.hpp"

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{
#endif /* !RECLS_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

class ReclsFileSearchDirectoryNode
    : public ReclsSearchDirectoryNode
{
public:
    typedef ReclsFileSearchDirectoryNode                                        class_type;
    typedef ::platformstl::filesystem_traits<recls_char_t>                      traits_type;
    typedef ::platformstl::basic_file_path_buffer<recls_char_t>                 file_path_buffer_type;
private:
    /// The string type
    typedef RECLS_STRING_TEMPLATE_1(recls_char_t)                               string_type;

    /// The file entry search type
#if defined(RECLS_PLATFORM_IS_UNIX)
    typedef unixstl_ns_qual(glob_sequence)                                      file_find_sequence_type;
#elif defined(RECLS_PLATFORM_IS_WIN32)
    typedef winstl_ns_qual(basic_findfile_sequence)<recls_char_t, traits_type>  file_find_sequence_type;
#endif /* platform */

    /// The directory search type
#if defined(RECLS_PLATFORM_IS_UNIX)
    typedef unixstl_ns_qual(readdir_sequence)                                   directory_sequence_type;
#elif defined(RECLS_PLATFORM_IS_WIN32)
    typedef file_find_sequence_type                                             directory_sequence_type;
#endif /* platform */

    /// The file entry search type
#ifdef _RECLS_USING_STLSOFT_SEARCHSPEC_SEQUENCE
    typedef stlsoft_ns_qual(searchspec_sequence)<file_find_sequence_type>       entry_sequence_type;
#else /* ? _RECLS_USING_STLSOFT_SEARCHSPEC_SEQUENCE */
    typedef file_find_sequence_type                                             entry_sequence_type;
#endif /* _RECLS_USING_STLSOFT_SEARCHSPEC_SEQUENCE */

// Construction
protected: // Not private, or GCC whines
    ReclsFileSearchDirectoryNode(   recls_uint32_t              flags
                                ,   size_t                      rootDirLen
                                ,   recls_char_t const          *searchDir
                                ,   recls_char_t const          *pattern
                                ,   hrecls_progress_fn_t        pfn
                                ,   recls_process_fn_param_t    param);
public:
    virtual ~ReclsFileSearchDirectoryNode();

    static class_type *FindAndCreate(recls_uint32_t             flags
                                ,   size_t                      rootDirLen
                                ,   recls_char_t const          *searchDir
                                ,   recls_char_t const          *pattern
                                ,   hrecls_progress_fn_t        pfn
                                ,   recls_process_fn_param_t    param);

    static recls_rc_t Stat( recls_char_t const  *path
                        ,   recls_uint32_t      flags
                        ,   recls_info_t        *phEntry);

// ReclsSearchDirectoryNode methods
private:
    /* virtual */ recls_rc_t GetNext();
    /* virtual */ recls_rc_t GetDetails(recls_info_t *pinfo);
    /* virtual */ recls_rc_t GetNextDetails(recls_info_t *pinfo);

// Implementation
private:
    recls_rc_t      Initialise();

#ifdef RECLS_ENFORCING_CONTRACTS
    recls_bool_t    is_valid() const;
#endif /* RECLS_ENFORCING_CONTRACTS */

    /// Translates the recls flags into the entry sequence flags
    static int essFlags_from_reclsFlags_(recls_uint32_t flags);
    /// Translates the recls flags into the directory sequence flags
    static int dssFlags_from_reclsFlags_(recls_uint32_t flags);
    /// Selects one or the other iterator, depending on the truth of b
    ///
    /// \note This is necessary because Borland seems not to like the tertiary 
    /// operator in the member initialiser list (MIL)
    static directory_sequence_type::const_iterator select_iter_if_( unsigned long                           b
                                                                ,   directory_sequence_type::const_iterator trueVal
                                                                ,   directory_sequence_type::const_iterator falseVal);

    /// Copies searchDir to buff, ensuring it has a trailing path name
    /// separator, and returns the precise length of the resultant string
    static size_t prepare_searchDir_(file_path_buffer_type &buff, recls_char_t const *searchDir);

    static recls_info_t     CreateEntryInfo(size_t                              rootDirLen
                                        ,   recls_char_t const                  *searchDir
                                        ,   size_t                              searchDirLen
                                        ,   recls_uint32_t                      flags
                                        ,   entry_sequence_type::const_iterator it);

// Members
private:
    recls_info_t                            m_current;
    class_type                              *m_dnode;
    const recls_uint32_t                    m_flags;
    const size_t                            m_rootDirLen;
    file_path_buffer_type                   m_searchDir;
    const size_t                            m_searchDirLen;
    const string_type                       m_pattern;
    directory_sequence_type                 m_directories;
    directory_sequence_type::const_iterator m_directoriesBegin;
    entry_sequence_type                     m_entries;
    entry_sequence_type::const_iterator     m_entriesBegin;
    const hrecls_progress_fn_t              m_pfn;
    const recls_process_fn_param_t          m_param;

// Not to be implemented
private:
    ReclsFileSearchDirectoryNode(class_type const &);
    class_type &operator =(class_type const &);
};

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */
