/* /////////////////////////////////////////////////////////////////////////////
 * File:        recls_debug.h
 *
 * Purpose:     Debug tracing.
 *
 * Created:     30th September 2003
 * Updated:     24th December 2005
 *
 * Home:        http://recls.org/
 *
 * Copyright (c) 2003-2005, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted in accordance with the license and warranty
 * information described in recls.h (included in this distribution, or available
 * from http://recls.org/)
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef RECLS_INCL_H_RECLS_DEBUG
#define RECLS_INCL_H_RECLS_DEBUG

/* File version */
#ifndef RECLS_DOCUMENTATION_SKIP_SECTION
# define RECLS_VER_H_RECLS_DEBUG_MAJOR      3
# define RECLS_VER_H_RECLS_DEBUG_MINOR      0
# define RECLS_VER_H_RECLS_DEBUG_REVISION   1
# define RECLS_VER_H_RECLS_DEBUG_EDIT       27
#endif /* !RECLS_DOCUMENTATION_SKIP_SECTION */

/** \file recls_debug.h Debug tracing
 */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

/* recls Header Files */
#include <recls/recls.h>
#include "recls_impl.h"

/* STLSoft / platform-specific Header Files */
#if defined(_RECLS_DEBUG)
# if defined(RECLS_PLATFORM_IS_UNIX)
#  include <stlsoft/errno_scope.hpp>
#  if defined(_RECLS_MT)
#   include <unixstl/tss_index.hpp>
#  endif /* _RECLS_MT */
# elif defined(RECLS_PLATFORM_IS_WIN32)
#  include <winstl/last_error_scope.hpp>
#   if defined(_RECLS_MT)
#    include <winstl/tss_index.hpp>
#   endif /* _RECLS_MT */
# else /* ? platform */
#  error Unrecognised platform
# endif /* platform */
# include <stdarg.h>
#endif /* _RECLS_DEBUG */

/* Standard C Header Files */
#include <stdio.h>
#if defined(_RECLS_DEBUG)
# include <string.h>
#endif /* _RECLS_DEBUG */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
namespace recls
{
#endif /* !RECLS_NO_NAMESPACE */

#if defined(_RECLS_DEBUG)
# if defined(RECLS_PLATFORM_IS_UNIX)
typedef stlsoft_ns_qual(errno_scope)        error_scope_t;
#  if defined(WIN32) && \
      defined(EMULATE_UNIX_ON_WIN32)
#  else /* ? WIN32 */
/* Really UNIX */
inline void OutputDebugStringA(char const *s)
{
    static_cast<void>(fprintf(stderr, s));
}
#  endif /* ? WIN32 */
#  if defined(_RECLS_MT)
unixstl_ns_using(tss_index)
#  endif /* _RECLS_MT */
# elif defined(RECLS_PLATFORM_IS_WIN32)
typedef winstl_ns_qual(last_error_scope)    error_scope_t;
#  if defined(_RECLS_MT)
winstl_ns_using(tss_index)
#  endif /* _RECLS_MT */
# else /* ? platform */
#  error Unrecognised platform
# endif /* platform */
#endif /* _RECLS_DEBUG */

/* /////////////////////////////////////////////////////////////////////////////
 * debug_printf
 */

#if defined(_RECLS_DEBUG)
inline void debug_printf(char const *fmt, ...)
{
    va_list args;
    char    _sz[2048];

    va_start(args, fmt);

# if defined(RECLS_PLATFORM_IS_UNIX) && \
     (   !defined(WIN32) || \
         !defined(EMULATE_UNIX_ON_WIN32))
    vsnprintf(_sz, RECLS_NUM_ELEMENTS(_sz), fmt, args);
# else /* ? OS */
    _vsnprintf(_sz, RECLS_NUM_ELEMENTS(_sz), fmt, args);
# endif /* OS */
    OutputDebugStringA(_sz);

    va_end(args);
}

class function_scope
{
public:
    function_scope(char const *fn)
    {
        error_scope_t    error_scope;

        ::strncpy(m_fn, fn, RECLS_NUM_ELEMENTS(m_fn) - 1);
        debug_printf("%*s>> %s()\n", static_cast<int>(post_inc_()), "", m_fn);
    }
    ~function_scope()
    {
        error_scope_t    error_scope;

        debug_printf("%*s<< %s()\n", static_cast<int>(pre_dec_()), "", m_fn);
    }

private:
    typedef ::stlsoft::sint32_t int32_t;

    union U
    {
    public:
        int32_t i;
        void    *pv;
    };

# if defined(_RECLS_MT)

#  if defined(RECLS_COMPILER_IS_INTEL)
#   pragma warning(push)
#   pragma warning(disable : 171)
#  endif /* RECLS_COMPILER_IS_INTEL */

    static int32_t tls_get_value_()
    {
        U u;

        u.pv = get_index_().get_value();

        return u.i;
    }
    static void tls_set_value_(int32_t i)
    {
        U u;

        u.i = i;

        get_index_().set_value(u.pv);
    }

#  if defined(RECLS_COMPILER_IS_INTEL)
#   pragma warning(pop)
#  endif /* RECLS_COMPILER_IS_INTEL */

# endif /* _RECLS_MT */

    static int32_t  post_inc_()
    {
# if defined(_RECLS_MT)
        int32_t i = (int32_t)(tls_get_value_());
        int32_t r = i++;

        tls_set_value_(i);

        return r;
# else /* ? _RECLS_MT */
        return get_index_()++;
# endif /* _RECLS_MT */
    }
    static int32_t  pre_dec_()
    {
# if defined(_RECLS_MT)
        int32_t i = (int32_t)(tls_get_value_());
        int32_t r = --i;

        tls_set_value_(i);

        return r;
# else /* ? _RECLS_MT */
        return --get_index_();
# endif /* _RECLS_MT */
    }

private:
    char                m_fn[1024];
# if defined(_RECLS_MT)
    static tss_index &get_index_()
    {
        static tss_index    s_index;

        return s_index;
    }
# else /* ? _RECLS_MT */
    static int32_t &get_index_()
    {
        static int32_t  s_count;

        return s_count;
    }
# endif /* _RECLS_MT */
};

# define function_scope_trace(f)        function_scope  _scope_ ## __LINE__(f)

#else /* ? _RECLS_DEBUG */

# ifdef __cplusplus
inline 
# else /* ? __cplusplus */
static 
# endif /* __cplusplus */
       void debug_printf_(char const *fmt, ...)
{
    static_cast<void>(fmt);
}
# define debug_printf                   (0) ? _RECLS_STATIC_CAST(void, 0) : debug_printf_
# define function_scope_trace(f)        do { ; } while(0)
#endif /* _RECLS_DEBUG */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#if !defined(RECLS_NO_NAMESPACE)
} /* namespace recls */
#endif /* !RECLS_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !RECLS_INCL_H_RECLS_DEBUG */

/* ////////////////////////////////////////////////////////////////////////// */
